""" Methods for generating new random starting candidates
for bulk crystal structures. If you find this implementation
useful in your work, please consider citing:
    M. Van den Bossche, Henrik Gronbeck, B. Hammer,
    J. Chem. Theory Comput., doi:10.1021/acs.jctc.8b00039
in addition to the papers mentioned in the docstrings."""
from random import shuffle, random, sample
import numpy as np
from ase import Atoms
from ase.data import atomic_numbers
from ase.build import molecule
from ase.ga.utilities import closest_distances_generator, atoms_too_close


def random_pos(cell):
    """ Returns a random position within the box
        described by the input box. """
    pos = np.zeros(3)
    for i in range(3):
        pos += random() * cell[i, :]
    return pos


class StartGenerator(object):
    """ Class for generating random starting candidates of bulk
    structures. The candidates are generated by:

      * Randomly creating a cell that satisfies the specified volume.
      * Adding one building block at a time, respecting the minimal
        distances. Building blocks can be either single atoms or
        groups of atoms.

    Parameters:

    blocks: list
            List of building units for the structure, each item
            as a (A, B) tuple, with A either an Atoms object or
            string (element symbol or molecule name) and B
            the number of these repeating units. A few examples:

            >>> sg = StartGenerator([('Ti', 4), ('O', 8)], ...)
            >>> sg = StartGenerator([('CO2', 10)], ...)
            >>> h2 = Atoms('H2', positions=[[0, 0, 0], [0.7, 0, 0]])
            >>> sg = StartGenerator([(h2, 12)], ...)

    blmin: dict or int
           Dictionary with minimal interatomic distances. If
           an integer is provided instead, the dictionary will be
           generated with this fraction of covalent bond lengths.

    volume: int or float
            Initial guess for the unit cell volume.

    cellbounds: ase.ga.bulk_utilities.CellBounds instance
                Describing limits on the cell shape, see
                :class:`~ase.ga.bulk_utilities.CellBounds`.

    cell: 3x3 matrix or length 3 or 6 vector
          If you want to keep the cell vectors fixed, the desired
          cell is to be specified here.

    splits: dict
            Splitting scheme to be used, based on:

            * `Lyakhov, Oganov, Valle, Comp. Phys. Comm. 181 (2010) 1623-32`__

            __ http://dx.doi.org/10.1016/j.cpc.2010.06.007

            This should be a dict specifying the relative probabilities
            for each split, written as tuples. For example,

            >>> splits = {(2,): 3, (1,): 1}

            This means that, for each structure, either a splitting
            factor of 2 is applied to one randomly chosen axis,
            or a splitting factor of 1 is applied (i.e., no splitting).
            The probability ratio of the two scenararios will be 3:1,
            i.e. 75% chance for the former and 25% chance for the latter
            splitting scheme.

            To e.g. always apply splitting factors of 2 and 3 along two
            randomly chosen axes:

            >>> splits={(2, 3): 1}
    """

    def __init__(self, blocks, blmin, volume, cellbounds=None,
                 cell=None, splits={(1,): 1}):

        self.volume = volume
        self.cellbounds = cellbounds
        self.cell = cell

        # normalize splitting probabilities:
        tot = sum([v for v in splits.values()])
        self.splits = {k: v * 1. / tot for k, v in splits.items()}

        # pre-process blocks and blmin:
        self.blocks = []
        numbers = []
        for block, count in blocks:
            if isinstance(block, Atoms):
                pass
            elif block in atomic_numbers:
                block = Atoms(block)
            else:
                block = molecule(block)
            self.blocks.append((block, count))
            numbers.extend(block.get_atomic_numbers())
        numbers = np.unique(numbers)

        if type(blmin) == dict:
            self.blmin = blmin
        else:
            self.blmin = closest_distances_generator(numbers, blmin)

    def get_new_candidate(self, maxiter=None):
        """ Returns a new candidate.

        maxiter: upper bound on the total number of times
                 the random position generator is called
                 when generating the new candidate.

                 By default (maxiter=None) no such bound
                 is imposed. If the generator takes too
                 long time to create a new candidate, it
                 may be suitable to specify a finite value.
                 When the bound is exceeded, None is returned.
        """

        pbc = [True] * 3
        blmin = self.blmin

        # generating the cell
        # cell splitting:
        # choose factors according to the probabilities
        r = random()
        cumprob = 0
        for split, prob in self.splits.items():
            cumprob += prob
            if cumprob > r:
                break

        directions = sample(range(3), len(split))
        repeat = [1, 1, 1]
        for i, d in enumerate(directions):
            repeat[d] = split[i]
        repeat = tuple(repeat)

        nparts = np.product(repeat)
        target_volume = self.volume / nparts

        # Randomly create a cell; without loss of generality,
        # a lower triangular form can be used, with tilt factors
        # within certain bounds.
        # For a cell to be valid, the full cell has to satisfy
        # the cellbounds constraints. Additionally, the length of
        # each subcell vector has to be greater than the largest
        # (X,X)-minimal-interatomic-distance in blmin.

        if self.cell is not None:
            full_cell = np.copy(self.cell)
            cell = (full_cell.T / repeat).T
            valid = True
        else:
            valid = False

        while not valid:
            blminmax = max([blmin[k] for k in blmin if k[0] == k[1]])
            cell = np.zeros((3, 3))
            l = target_volume**0.33
            cell[0, 0] = random() * l
            cell[1, 0] = (random() - 0.5) * cell[0, 0]
            cell[1, 1] = random() * l
            cell[2, 0] = (random() - 0.5) * cell[0, 0]
            cell[2, 1] = (random() - 0.5) * cell[1, 1]
            cell[2, 2] = random() * l

            volume = abs(np.linalg.det(cell))
            cell *= (target_volume / volume)**0.33

            full_cell = (repeat * cell.T).T

            valid = True
            if self.cellbounds is not None:
                if not self.cellbounds.is_within_bounds(full_cell):
                    valid = False
            for i in range(3):
                if np.linalg.norm(cell[i, :]) < blminmax:
                    valid = False

        # generating the atomic positions
        blocks = []
        surplus = []
        indices = []
        for i, (block, count) in enumerate(self.blocks):
            count_part = int(np.ceil(count * 1. / nparts))
            surplus.append(nparts * count_part - count)
            blocks.extend([block] * count_part)
            indices.extend([i] * count_part)

        N_blocks = len(blocks)

        # The ordering is shuffled so different blocks
        # are added in random order.
        order = list(range(N_blocks))
        shuffle(order)
        blocks = [blocks[i] for i in order]
        indices = np.array(indices)[order]

        # Runs until we have found a valid candidate.
        cand = Atoms('', cell=cell, pbc=pbc)
        niter = 0
        for i in range(N_blocks):
            atoms = blocks[i].copy()
            atoms.set_tags(i)
            rotate = len(atoms) > 1

            # Make each new position one at a time.
            while maxiter is None or niter < maxiter:
                cop = atoms.get_positions().mean(axis=0)
                pos = random_pos(cell)
                atoms.translate(pos - cop)
                if rotate:
                    phi, theta, psi = 360 * np.random.random(3)
                    atoms.euler_rotate(phi=phi, theta=0.5 * theta, psi=psi,
                                       center=pos)
                # add if it fits:
                attempt = cand + atoms
                attempt.wrap()
                too_close = atoms_too_close(attempt, blmin, use_tags=True)
                if not too_close:
                    cand += atoms
                    break
                niter += 1
            else:
                # Reached upper bound on iteration count
                cand = None
                break

        if cand is None:
            return None

        # rebuild the candidate after repeating,
        # randomly deleting surplus blocks and
        # sorting back to the original order
        tags = cand.get_tags()
        nrep = int(np.prod(repeat))
        cand_full = cand.repeat(repeat)

        tags_full = cand_full.get_tags()
        for i in range(nrep):
            tags_full[len(cand) * i:len(cand) * (i + 1)] += i * N_blocks
        cand_full.set_tags(tags_full)

        cand = Atoms('', cell=full_cell, pbc=pbc)
        indices_full = np.tile(indices, nrep)
        tag_counter = 0
        for i, (block, count) in enumerate(self.blocks):
            tags = np.where(indices_full == i)[0]
            bad = np.random.choice(tags, size=surplus[i], replace=False)
            for tag in tags:
                if tag not in bad:
                    select = [a.index for a in cand_full if a.tag == tag]
                    atoms = cand_full[select]  # is indeed a copy!
                    atoms.set_tags(tag_counter)
                    assert len(atoms) == len(block)
                    cand += atoms
                    tag_counter += 1

        return cand



"""Test that atoms.get_center_of_mass(scaled=True) works"""

import numpy as np
from ase import Atoms

d = 1.142
a = Atoms('CO', positions=[(2, 0, 0), (2, -d, 0)], pbc=True)
a.set_cell(np.array(((4, -4, 0), (0, 5.657, 0), (0, 0, 10))))

def array_almost_equal(a1, a2, tol=np.finfo(type(1.0)).eps):
    return (np.abs(a1 - a2) < tol).all()

scaledref = np.array((0.5, 0.23823622, 0.))
assert array_almost_equal(a.get_center_of_mass(scaled=True), scaledref, tol=1e-8)



import numpy as np
from ase.build import bulk
from ase.calculators.emt import EMT

a = bulk('Cu', 'fcc')
a.calc = EMT()
a.set_cell(np.dot(a.cell,
                  [[1.02, 0, 0.03],
                   [0, 0.99, -0.02],
                   [0.1, -0.01, 1.03]]),
           scale_atoms=True)
a *= (1, 2, 3)
a.rattle()
# Verify analytical stress tensor against numerical value
s_analytical = a.get_stress()
s_numerical = a.calc.calculate_numerical_stress(a, 1e-5)
s_p_err = 100 * (s_numerical - s_analytical) / s_numerical
print('Analytical stress:\n', s_analytical)
print('Numerical stress:\n', s_numerical)
print('Percent error in stress:\n', s_p_err)
assert np.all(abs(s_p_err) < 1e-5)



# creates: zno.png
import numpy as np
import matplotlib.pyplot as plt
from ase.phasediagram import Pourbaix, solvated
refs = solvated('Zn')
print(refs)
refs += [('Zn', 0.0), ('ZnO', -3.323), ('ZnO2(aq)', -2.921)]
pb = Pourbaix(refs, Zn=1, O=1)
print(pb.decompose(1.0, 9.0))
U = np.linspace(-2, 2, 200)
pH = np.linspace(-2, 16, 300)
d, names, text = pb.diagram(U, pH, plot=True, show=False)
print(names, text)
assert len(names) == 7
plt.savefig('zno.png')




# lammps.py (2011/03/29)
# An ASE calculator for the LAMMPS classical MD code available from
#       http://lammps.sandia.gov/
# The environment variable ASE_LAMMPSRUN_COMMAND must be defined to point to the
# LAMMPS binary.
#
# Copyright (C) 2009 - 2011 Joerg Meyer, joerg.meyer@ch.tum.de
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this file; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA or see <http://www.gnu.org/licenses/>.


import os
import shutil
import shlex
from subprocess import Popen, PIPE
from threading import Thread
from re import compile as re_compile, IGNORECASE
from tempfile import mkdtemp, NamedTemporaryFile, mktemp as uns_mktemp
import inspect
import warnings
import numpy as np

from ase import Atoms
from ase.parallel import paropen
from ase.calculators.calculator import Calculator
from ase.calculators.calculator import all_changes
from ase.utils import basestring as asestring
from ase.data import chemical_symbols
from ase.data import atomic_masses
from ase.io.lammpsdata import write_lammps_data
from ase.io.lammpsrun import read_lammps_dump
from ase.calculators.lammps import Prism
from ase.calculators.lammps import write_lammps_in
from ase.calculators.lammps import CALCULATION_END_MARK
from ase.calculators.lammps import convert

__all__ = ["LAMMPS"]


class LAMMPS(Calculator):
    """The LAMMPS calculators object

    files: list
        List of files typically containing relevant potentials for the calculation
    parameters: dict
        Dictionary of settings to be passed into the input file for calculation.
    specorder: list
        Within LAAMPS, atoms are identified by an integer value starting from 1.
        This variable allows the user to define the order of the indices assigned to the
        atoms in the calculation, with the default if not given being alphabetical
    keep_tmp_files: bool
        Retain any temporary files created. Mostly useful for debugging.
    tmp_dir: str
        path/dirname (default None -> create automatically).
        Explicitly control where the calculator object should create
        its files. Using this option implies 'keep_tmp_files'
    no_data_file: bool
        Controls whether an explicit data file will be used for feeding
        atom coordinates into lammps. Enable it to lessen the pressure on
        the (tmp) file system. THIS OPTION MIGHT BE UNRELIABLE FOR CERTAIN
        CORNER CASES (however, if it fails, you will notice...).
    keep_alive: bool
        When using LAMMPS as a spawned subprocess, keep the subprocess
        alive (but idling when unused) along with the calculator object.
    always_triclinic: bool
        Force use of a triclinic cell in LAMMPS, even if the cell is
        a perfect parallelepiped.

        **Example**

Provided that the respective potential file is in the working directory, one
can simply run (note that LAMMPS needs to be compiled to work with EAM
potentials)

::

    from ase import Atom, Atoms
    from ase.build import bulk
    from ase.calculators.lammpsrun import LAMMPS

    parameters = {'pair_style': 'eam/alloy',
                  'pair_coeff': ['* * NiAlH_jea.eam.alloy H Ni']}

    files = ['NiAlH_jea.eam.alloy']

    Ni = bulk('Ni', cubic=True)
    H = Atom('H', position=Ni.cell.diagonal()/2)
    NiH = Ni + H

    lammps = LAMMPS(parameters=parameters, files=files)

    NiH.set_calculator(lammps)
    print("Energy ", NiH.get_potential_energy())

(Remember you also need to set the environment variable ``$ASE_LAMMPSRUN_COMMAND``)

    """

    name = "lammpsrun"
    implemented_properties = ["energy", "forces", "stress", "energies"]

    # parameters to choose options in LAMMPSRUN
    ase_parameters = dict(
        specorder=None,
        always_triclinic=False,
        keep_alive=True,
        keep_tmp_files=False,
        no_data_file=False,
        tmp_dir=None,
        files=[],  # usually contains potential parameters
        verbose=False,
        write_velocities=False,
        binary_dump=True,  # bool - use binary dump files (full
                           # precision but long long ids are casted to
                           # double)
        lammps_options="-echo log -screen none -log /dev/stdout",
        trajectory_out=None,  # file object, if is not None the
                              # trajectory will be saved in it
    )

    # parameters forwarded to LAMMPS
    lammps_parameters = dict(
        boundary=None,  # bounadry conditions styles
        units="metal",  # str - Which units used; some potentials
                        # require certain units
        atom_style="atomic",
        special_bonds=None,
        # potential informations
        pair_style="lj/cut 2.5",
        pair_coeff=["* * 1 1"],
        masses=None,
        pair_modify=None,
        # variables controlling the output
        thermo_args=[
            "step", "temp", "press", "cpu", "pxx", "pyy", "pzz",
            "pxy", "pxz", "pyz", "ke", "pe", "etotal", "vol", "lx",
            "ly", "lz", "atoms", ],
        dump_properties=["id", "type", "x", "y", "z", "vx", "vy",
                         "vz", "fx", "fy", "fz", ],
        dump_period=1,  # period of system snapshot saving (in MD steps)
    )

    default_parameters = dict(ase_parameters, **lammps_parameters)

    # legacy parameter persist, when the 'parameters' dictinary is manipulated
    # from the outside.  All others are rested to the default value
    legacy_parameters = [
        "specorder",
        "dump_period",
        "always_triclinic",
        "keep_alive",
        "keep_tmp_files",
        "tmp_dir",
        "parameters",
        "no_data_file",
        "files",
        "write_velocities",
        "trajectory_out",
    ]

    legacy_parameters_map = {"_custom_thermo_args": "thermo_args"}

    legacy_warn_string = "You are using an "
    legacy_warn_string += "old syntax to set '{}'.\n"
    legacy_warn_string += "Please use {}.set().".format(name.upper())

    def __init__(self, label="lammps", **kwargs):
        # "Parameters" used to be the dictionary with all parameters forwarded
        # to lammps.  This clashes with the implementation in Calculator to
        # reload an old one. Trying to catch both cases to not break old
        # scripts.
        if "parameters" in kwargs:
            old_parameters = kwargs["parameters"]
            if isinstance(old_parameters, dict):
                warnings.warn(self.legacy_warn_string.format("parameters"))
                del kwargs["parameters"]
        else:
            old_parameters = None

        Calculator.__init__(self, label=label, **kwargs)

        if old_parameters and isinstance(old_parameters, dict):
            self.set(**old_parameters)

        self.prism = None
        self.calls = 0
        self.forces = None
        # thermo_content contains data "written by" thermo_style.
        # It is a list of dictionaries, each dict (one for each line
        # printed by thermo_style) contains a mapping between each
        # custom_thermo_args-argument and the corresponding
        # value as printed by lammps. thermo_content will be
        # re-populated by the read_log method.
        self.thermo_content = []

        if self.parameters.tmp_dir is not None:
            # If tmp_dir is pointing somewhere, don't remove stuff!
            self.parameters.keep_tmp_files = True
        self._lmp_handle = None  # To handle the lmp process

        if self.parameters.tmp_dir is None:
            self.parameters.tmp_dir = mkdtemp(prefix="LAMMPS-")
        else:
            self.parameters.tmp_dir = os.path.realpath(self.parameters.tmp_dir)
            if not os.path.isdir(self.parameters.tmp_dir):
                os.mkdir(self.parameters.tmp_dir, 0o755)

        for f in self.parameters.files:
            shutil.copy(
                f, os.path.join(self.parameters.tmp_dir, os.path.basename(f))
            )

    def get_lammps_command(self):
        cmd = self.parameters.get('command')
        if cmd is None:
            envvar = 'ASE_{}_COMMAND'.format(self.name.upper())
            cmd = os.environ.get(envvar)

        if cmd is None:
            cmd = 'lammps'

        opts = self.parameters.get('lammps_options')

        if opts is not None:
            cmd = '{} {}'.format(cmd, opts)

        return cmd

    def __setattr__(self, key, value):
        """Catch attribute sets to emulate legacy behavior.

        Old LAMMPSRUN allows to just override the parameters
        dictionary. "Modern" ase calculators can assume that default
        parameters are always set, overrides of the
        'parameters'-dictionary have to be caught and the default
        parameters need to be added first.  A check refuses to set
        calculator attributes if they are unknown and set outside the
        '__init__' functions.
        """
        # !TODO: remove and break somebody's code (e.g. the test examples)
        if (
                key == "parameters"
                and value is not None
                and self.parameters is not None
        ):
            temp_dict = self.get_default_parameters()
            if self.parameters:
                for l_key in self.legacy_parameters:
                    try:
                        temp_dict[l_key] = self.parameters[l_key]
                    except KeyError:
                        pass
            temp_dict.update(value)
            value = temp_dict
        if key in self.legacy_parameters and key != "parameters":
            warnings.warn(self.legacy_warn_string.format(key))
            self.set(**{key: value})
        elif key in self.legacy_parameters_map:
            warnings.warn(
                self.legacy_warn_string.format(
                    "{} for {}".format(self.legacy_parameters_map[key], key)
                )
            )
            self.set(**{self.legacy_parameters_map[key]: value})
        # Catch setting none-default attributes
        # one test was assigning an useless Attribute, but it still worked
        # because the assigned object was before manipulation already handed
        # over to the calculator (10/2018)
        elif hasattr(self, key) or inspect.stack()[1][3] == "__init__":
            Calculator.__setattr__(self, key, value)
        else:
            raise AttributeError("Setting unknown Attribute '{}'".format(key))

    def __getattr__(self, key):
        """Corresponding getattribute-function to emulate legacy behavior.
        """
        if key in self.legacy_parameters and key != "parameters":
            return self.parameters[key]
        if key in self.legacy_parameters_map:
            return self.parameters[self.legacy_parameters_map[key]]
        return object.__getattribute__(self, key)

    def clean(self, force=False):

        self._lmp_end()

        if not self.parameters.keep_tmp_files or force:
            shutil.rmtree(self.parameters.tmp_dir)

    def check_state(self, atoms, tol=1.0e-10):
        # Transforming the unit cell to conform to LAMMPS' convention for
        # orientation (c.f. https://lammps.sandia.gov/doc/Howto_triclinic.html)
        # results in some precision loss, so we use bit larger tolerance than
        # machine precision here.  Note that there can also be precision loss
        # related to how many significant digits are specified for things in
        # the LAMMPS input file.
        return Calculator.check_state(self, atoms, tol)

    def calculate(self, atoms=None, properties=None, system_changes=None):
        if properties is None:
            properties = self.implemented_properties
        if system_changes is None:
            system_changes = all_changes
        Calculator.calculate(self, atoms, properties, system_changes)
        self.run()

    def _lmp_alive(self):
        # Return True if this calculator is currently handling a running
        # lammps process
        return self._lmp_handle and not isinstance(
            self._lmp_handle.poll(), int
        )

    def _lmp_end(self):
        # Close lammps input and wait for lammps to end. Return process
        # return value
        if self._lmp_alive():
            self._lmp_handle.stdin.close()
            # !TODO: handle lammps error codes
            # return self._lmp_handle.wait()

    def set_missing_parameters(self):
        """Verify that all necessary variables are set.
        """
        symbols = self.atoms.get_chemical_symbols()
        # If unspecified default to atom types in alphabetic order
        if not self.parameters.specorder:
            self.parameters.specorder = sorted(set(symbols))

        # !TODO: handle cases were setting masses actual lead to errors
        if not self.parameters.masses:
            self.parameters.masses = []
            for type_id, specie in enumerate(self.parameters.specorder):
                mass = atomic_masses[chemical_symbols.index(specie)]
                self.parameters.masses += [
                    "{0:d} {1:f}".format(type_id + 1, mass)
                ]

        # set boundary condtions
        if not self.parameters.boundary:
            b_str = " ".join(["fp"[int(x)] for x in self.atoms.get_pbc()])
            self.parameters.boundary = b_str

    def run(self, set_atoms=False):
        # !TODO: split this function
        """Method which explicitly runs LAMMPS."""
        pbc = self.atoms.get_pbc()
        if all(pbc):
            cell = self.atoms.get_cell()
        elif not any(pbc):
            # large enough cell for non-periodic calculation -
            # LAMMPS shrink-wraps automatically via input command
            #       "periodic s s s"
            # below
            cell = 2 * np.max(np.abs(self.atoms.get_positions())) * np.eye(3)
        else:
            warnings.warn(
                "semi-periodic ASE cell detected - translation "
                + "to proper LAMMPS input cell might fail"
            )
            cell = self.atoms.get_cell()
        self.prism = Prism(cell)

        self.set_missing_parameters()
        self.calls += 1

        # change into subdirectory for LAMMPS calculations
        cwd = os.getcwd()
        os.chdir(self.parameters.tmp_dir)

        # setup file names for LAMMPS calculation
        label = "{0}{1:>06}".format(self.label, self.calls)
        lammps_in = uns_mktemp(
            prefix="in_" + label, dir=self.parameters.tmp_dir
        )
        lammps_log = uns_mktemp(
            prefix="log_" + label, dir=self.parameters.tmp_dir
        )
        lammps_trj_fd = NamedTemporaryFile(
            prefix="trj_" + label,
            suffix=(".bin" if self.parameters.binary_dump else ""),
            dir=self.parameters.tmp_dir,
            delete=(not self.parameters.keep_tmp_files),
        )
        lammps_trj = lammps_trj_fd.name
        if self.parameters.no_data_file:
            lammps_data = None
        else:
            lammps_data_fd = NamedTemporaryFile(
                prefix="data_" + label,
                dir=self.parameters.tmp_dir,
                delete=(not self.parameters.keep_tmp_files),
                mode='w',
                encoding='ascii'
            )
            write_lammps_data(
                lammps_data_fd,
                self.atoms,
                specorder=self.parameters.specorder,
                force_skew=self.parameters.always_triclinic,
                velocities=self.parameters.write_velocities,
                prismobj=self.prism,
                units=self.parameters.units,
                atom_style=self.parameters.atom_style
            )
            lammps_data = lammps_data_fd.name
            lammps_data_fd.flush()

        # see to it that LAMMPS is started
        if not self._lmp_alive():
            command = self.get_lammps_command()
            # Attempt to (re)start lammps
            self._lmp_handle = Popen(
                shlex.split(command, posix=(os.name == "posix")),
                stdin=PIPE,
                stdout=PIPE,
            )
        lmp_handle = self._lmp_handle

        # Create thread reading lammps stdout (for reference, if requested,
        # also create lammps_log, although it is never used)
        if self.parameters.keep_tmp_files:
            lammps_log_fd = open(lammps_log, "wb")
            fd = SpecialTee(lmp_handle.stdout, lammps_log_fd)
        else:
            fd = lmp_handle.stdout
        thr_read_log = Thread(target=self.read_lammps_log, args=(fd,))
        thr_read_log.start()

        # write LAMMPS input (for reference, also create the file lammps_in,
        # although it is never used)
        if self.parameters.keep_tmp_files:
            lammps_in_fd = open(lammps_in, "wb")
            fd = SpecialTee(lmp_handle.stdin, lammps_in_fd)
        else:
            fd = lmp_handle.stdin
        write_lammps_in(
            lammps_in=fd,
            parameters=self.parameters,
            atoms=self.atoms,
            prismobj=self.prism,
            lammps_trj=lammps_trj,
            lammps_data=lammps_data,
        )

        if self.parameters.keep_tmp_files:
            lammps_in_fd.close()

        # Wait for log output to be read (i.e., for LAMMPS to finish)
        # and close the log file if there is one
        thr_read_log.join()
        if self.parameters.keep_tmp_files:
            lammps_log_fd.close()

        if not self.parameters.keep_alive:
            self._lmp_end()

        exitcode = lmp_handle.poll()
        if exitcode and exitcode != 0:
            cwd = os.getcwd()
            raise RuntimeError(
                "LAMMPS exited in {} with exit code: {}."
                "".format(cwd, exitcode)
            )

        # A few sanity checks
        if len(self.thermo_content) == 0:
            raise RuntimeError("Failed to retrieve any thermo_style-output")
        if int(self.thermo_content[-1]["atoms"]) != len(self.atoms):
            # This obviously shouldn't happen, but if prism.fold_...() fails,
            # it could
            raise RuntimeError("Atoms have gone missing")

        trj_atoms = read_lammps_dump(
            infileobj=lammps_trj,
            order=False,
            index=-1,
            prismobj=self.prism,
            specorder=self.parameters.specorder,
        )

        if set_atoms:
            self.atoms = trj_atoms.copy()

        self.forces = trj_atoms.get_forces()
        # !TODO: trj_atoms is only the last snapshot of the system; Is it
        #        desireable to save also the inbetween steps?
        if self.parameters.trajectory_out is not None:
            # !TODO: is it advisable to create here temporary atoms-objects
            self.trajectory_out.write(trj_atoms)

        tc = self.thermo_content[-1]
        self.results["energy"] = convert(
            tc["pe"], "energy", self.parameters["units"], "ASE"
        )
        self.results["free_energy"] = self.results["energy"]
        self.results["forces"] = self.forces.copy()
        stress = np.array(
            [-tc[i] for i in ("pxx", "pyy", "pzz", "pyz", "pxz", "pxy")]
        )

        # We need to apply the Lammps rotation stuff to the stress:
        xx, yy, zz, yz, xz, xy = stress
        stress_tensor = np.array([[xx, xy, xz],
                                  [xy, yy, yz],
                                  [xz, yz, zz]])
        R = self.prism.rot_mat
        stress_atoms = np.dot(R, stress_tensor)
        stress_atoms = np.dot(stress_atoms, R.T)
        stress_atoms = stress_atoms[[0, 1, 2, 1, 0, 0],
                                    [0, 1, 2, 2, 2, 1]]
        stress = stress_atoms

        self.results["stress"] = convert(
            stress, "pressure", self.parameters["units"], "ASE"
        )

        lammps_trj_fd.close()
        if not self.parameters.no_data_file:
            lammps_data_fd.close()

        os.chdir(cwd)

    def read_lammps_log(self, lammps_log=None):
        # !TODO: somehow communicate 'thermo_content' explicitly
        """Method which reads a LAMMPS output log file."""

        if lammps_log is None:
            lammps_log = self.label + ".log"

        if isinstance(lammps_log, asestring):
            fileobj = paropen(lammps_log, "wb")
            close_log_file = True
        else:
            # Expect lammps_in to be a file-like object
            fileobj = lammps_log
            close_log_file = False

        # read_log depends on that the first (three) thermo_style custom args
        # can be capitilized and matched against the log output. I.e.
        # don't use e.g. 'ke' or 'cpu' which are labeled KinEng and CPU.
        _custom_thermo_mark = " ".join(
            [x.capitalize() for x in self.parameters.thermo_args[0:3]]
        )

        # !TODO: regex-magic necessary?
        # Match something which can be converted to a float
        f_re = r"([+-]?(?:(?:\d+(?:\.\d*)?|\.\d+)(?:e[+-]?\d+)?|nan|inf))"
        n_args = len(self.parameters["thermo_args"])
        # Create a re matching exactly N white space separated floatish things
        _custom_thermo_re = re_compile(
            r"^\s*" + r"\s+".join([f_re] * n_args) + r"\s*$", flags=IGNORECASE
        )

        thermo_content = []
        line = fileobj.readline().decode("utf-8")
        while line and line.strip() != CALCULATION_END_MARK:
            # check error
            if 'ERROR:' in line:
                if close_log_file:
                    fileobj.close()
                raise RuntimeError('LAMMPS exits with error message: {}'.format(line))

            # get thermo output
            if line.startswith(_custom_thermo_mark):
                bool_match = True
                while bool_match:
                    line = fileobj.readline().decode("utf-8")
                    bool_match = _custom_thermo_re.match(line)
                    if bool_match:
                        # create a dictionary between each of the
                        # thermo_style args and it's corresponding value
                        thermo_content.append(
                            dict(
                                zip(
                                    self.parameters.thermo_args,
                                    map(float, bool_match.groups()),
                                )
                            )
                        )
            else:
                line = fileobj.readline().decode("utf-8")

        if close_log_file:
            fileobj.close()

        self.thermo_content = thermo_content


class SpecialTee:
    """A special purpose, with limited applicability, tee-like thing.

    A subset of stuff read from, or written to, orig_fd,
    is also written to out_fd.
    It is used by the lammps calculator for creating file-logs of stuff
    read from, or written to, stdin and stdout, respectively.
    """

    def __init__(self, orig_fd, out_fd):
        self._orig_fd = orig_fd
        self._out_fd = out_fd
        self.name = orig_fd.name

    def write(self, data):
        self._orig_fd.write(data)
        self._out_fd.write(data)
        self.flush()

    def read(self, *args, **kwargs):
        data = self._orig_fd.read(*args, **kwargs)
        self._out_fd.write(data)
        return data

    def readline(self, *args, **kwargs):
        data = self._orig_fd.readline(*args, **kwargs)
        self._out_fd.write(data)
        return data

    def readlines(self, *args, **kwargs):
        data = self._orig_fd.readlines(*args, **kwargs)
        self._out_fd.write("".join(data))
        return data

    def flush(self):
        self._orig_fd.flush()
        self._out_fd.flush()


if __name__ == "__main__":
    pair_style = "eam"
    Pd_eam_file = "Pd_u3.eam"
    pair_coeff = ["* * " + Pd_eam_file]
    parameters = {"pair_style": pair_style, "pair_coeff": pair_coeff}
    my_files = [Pd_eam_file]
    calc = LAMMPS(parameters=parameters, files=my_files)
    a0 = 3.93
    b0 = a0 / 2.0

    bulk = Atoms(
        ["Pd"] * 4,
        positions=[(0, 0, 0), (b0, b0, 0), (b0, 0, b0), (0, b0, b0)],
        cell=[a0] * 3,
        pbc=True,
    )
    # test get_forces
    print("forces for a = {0}".format(a0))
    print(calc.get_forces(bulk))
    # single points for various lattice constants
    bulk.set_calculator(calc)
    for i in range(-5, 5, 1):
        a = a0 * (1 + i / 100.0)
        bulk.set_cell([a] * 3)
        print("a : {0} , total energy : {1}".format(
            a, bulk.get_potential_energy()))

    calc.clean()



from gpaw import GPAW
calc = GPAW('bulk.Ag.gpw')
atoms = calc.get_atoms()
path = atoms.cell.bandpath('WLGXWK', density=10)
path.write('path.json')

calc.set(kpts=path, fixdensity=True, symmetry='off')

atoms.get_potential_energy()
bs = calc.band_structure()
bs.write('bs.json')



# Run flake8 on main source dir and documentation.
import sys
import unittest
from collections import defaultdict
from pathlib import Path
from subprocess import Popen, PIPE

import ase

try:
    import flake8  # noqa
except ImportError:
    raise unittest.SkipTest('flake8 module not available')


asepath = Path(ase.__path__[0])


def run_flake8():
    proc = Popen([sys.executable,
                  '-m',
                  'flake8',
                  str(asepath),
                  str((asepath / '../doc').resolve()),
                  '--exclude',
                  str((asepath / '../doc/build/*').resolve()),
                  '--ignore',
                  'E129,W293,W503,W504,E741'
                  '-j',
                  '1'],
                 stdout=PIPE)
    stdout, stderr = proc.communicate()
    stdout = stdout.decode('utf8')

    errors = defaultdict(int)
    files = defaultdict(int)
    descriptions = {}
    for stdout_line in stdout.splitlines():
        tokens = stdout_line.split(':', 3)
        filename, lineno, colno, complaint = tokens
        lineno = int(lineno)
        e = complaint.strip().split()[0]
        errors[e] += 1
        descriptions[e] = complaint
        files[filename] += 1

    print('Bad files:')
    for f, n in sorted(files.items(), key=lambda i: i[1]):
        print(f, n)

    print('\nmax_errors = {')
    for e, n in sorted(errors.items(), key=lambda i: i[1]):
        print('    # {}\n    {!r}: {},'
              .format(descriptions[e][6:], e, n))
    print('}')

    for e, n in errors.items():
        if n > max_errors.get(e, 0):
            raise ValueError(
                'Maximum number of flake8 errors exceeded: {} * {}.  '
                'Please run flake8 on your code and clean up.'
                .format(n, e))


max_errors = {
    # do not compare types, use 'isinstance()'
    'E721': 1,
    # multiple imports on one line
    'E401': 1,
    # multiple spaces before keyword
    'E272': 1,
    # continuation line under-indented for hanging indent
    'E121': 2,
    # whitespace before '('
    'E211': 2,
    # continuation line with same indent as next logical line
    'E125': 3,
    # comparison to True should be 'if cond is True:' or 'if cond:'
    'E712': 3,
    # no newline at end of file
    'W292': 3,
    # missing whitespace after keyword
    'E275': 3,
    # multiple spaces after operator
    'E222': 4,
    # missing whitespace around modulo operator
    'E228': 4,
    # expected 1 blank line before a nested definition, found 0
    'E306': 5,
    # test for membership should be 'not in'
    'E713': 4,
    # multiple statements on one line (colon)
    'E701': 5,
    # indentation is not a multiple of four (comment)
    'E114': 5,
    # unexpected indentation (comment)
    'E116': 5,
    # comparison to None should be 'if cond is None:'
    'E711': 5,
    # expected 1 blank line, found 0
    'E301': 8,
    # multiple spaces after keyword
    'E271': 8,
    # test for object identity should be 'is not'
    'E714': 8,
    # closing bracket does not match visual indentation
    'E124': 8,
    # too many leading '#' for block comment
    'E266': 10,
    # over-indented
    'E117': 11,
    # indentation contains mixed spaces and tabs
    'E101': 12,
    # indentation contains tabs
    'W191': 13,
    # closing bracket does not match indentation of opening bracket's line
    'E123': 14,
    # multiple spaces before operator
    'E221': 16,
    # whitespace before '}'
    'E202': 19,
    # whitespace after '{'
    'E201': 20,
    # inline comment should start with '# '
    'E262': 20,
    # the backslash is redundant between brackets
    'E502': 30,
    # continuation line missing indentation or outdented
    'E122': 31,
    # indentation is not a multiple of four
    'E111': 36,
    # do not use bare 'except'
    'E722': 38,
    # whitespace before ':'
    'E203': 38,
    # blank line at end of file
    'W391': 49,
    # continuation line over-indented for hanging indent
    'E126': 48,
    # multiple spaces after ','
    'E241': 50,
    # continuation line under-indented for visual indent
    'E128': 53,
    # continuation line over-indented for visual indent
    'E127': 60,
    # missing whitespace around operator
    'E225': 61,
    # ambiguous variable name 'O'
    'E741': 77,
    # too many blank lines (2)
    'E303': 82,
    # expected 2 blank lines after class or function definition, found 1
    'E305': 83,
    # module level import not at top of file
    'E402': 91,
    # at least two spaces before inline comment
    'E261': 97,
    # expected 2 blank lines, found 1
    'E302': 111,
    # unexpected spaces around keyword / parameter equals
    'E251': 117,
    # trailing whitespace
    'W291': 222,
    # block comment should start with '# '
    'E265': 246,
    # missing whitespace after ','
    'E231': 465,
    # missing whitespace around arithmetic operator
    'E226': 563,
    # line too long (93 > 79 characters)
    'E501': 737}

run_flake8()



import numpy.random as random
import numpy as np
from ase import Atoms
from ase.neighborlist import (NeighborList, PrimitiveNeighborList,
                              NewPrimitiveNeighborList)
from ase.build import bulk

atoms = Atoms(numbers=range(10),
              cell=[(0.2, 1.2, 1.4),
                    (1.4, 0.1, 1.6),
                    (1.3, 2.0, -0.1)])
atoms.set_scaled_positions(3 * random.random((10, 3)) - 1)


def count(nl, atoms):
    c = np.zeros(len(atoms), int)
    R = atoms.get_positions()
    cell = atoms.get_cell()
    d = 0.0
    for a in range(len(atoms)):
        i, offsets = nl.get_neighbors(a)
        for j in i:
            c[j] += 1
        c[a] += len(i)
        d += (((R[i] + np.dot(offsets, cell) - R[a])**2).sum(1)**0.5).sum()
    return d, c

for sorted in [False, True]:
    for p1 in range(2):
        for p2 in range(2):
            for p3 in range(2):
                # print(p1, p2, p3)
                atoms.set_pbc((p1, p2, p3))
                nl = NeighborList(atoms.numbers * 0.2 + 0.5,
                                  skin=0.0, sorted=sorted)
                nl.update(atoms)
                d, c = count(nl, atoms)
                atoms2 = atoms.repeat((p1 + 1, p2 + 1, p3 + 1))
                nl2 = NeighborList(atoms2.numbers * 0.2 + 0.5,
                                   skin=0.0, sorted=sorted)
                nl2.update(atoms2)
                d2, c2 = count(nl2, atoms2)
                c2.shape = (-1, 10)
                dd = d * (p1 + 1) * (p2 + 1) * (p3 + 1) - d2
                assert abs(dd) < 1e-10
                assert not (c2 - c).any()

h2 = Atoms('H2', positions=[(0, 0, 0), (0, 0, 1)])
nl = NeighborList([0.5, 0.5], skin=0.1, sorted=True, self_interaction=False)
nl2 = NeighborList([0.5, 0.5], skin=0.1, sorted=True, self_interaction=False, primitive=NewPrimitiveNeighborList)
assert nl2.update(h2)
assert nl.update(h2)
assert not nl.update(h2)
assert (nl.get_neighbors(0)[0] == [1]).all()
m = np.zeros((2,2))
m[0,1] = 1
assert np.array_equal(nl.get_connectivity_matrix(sparse=False), m)
assert np.array_equal(nl.get_connectivity_matrix(sparse=True).todense(), m)
assert np.array_equal(nl.get_connectivity_matrix().todense(), nl2.get_connectivity_matrix().todense())

h2[1].z += 0.09
assert not nl.update(h2)
assert (nl.get_neighbors(0)[0] == [1]).all()

h2[1].z += 0.09
assert nl.update(h2)
assert (nl.get_neighbors(0)[0] == []).all()
assert nl.nupdates == 2

h2 = Atoms('H2', positions=[(0, 0, 0), (0, 0, 1)])
nl = NeighborList([0.1, 0.1], skin=0.1, bothways=True, self_interaction=False)
assert nl.update(h2)
assert nl.get_neighbors(0)[1].shape == (0, 3)
assert nl.get_neighbors(0)[1].dtype == int

x = bulk('X', 'fcc', a=2**0.5)

nl = NeighborList([0.5], skin=0.01, bothways=True, self_interaction=False)
nl.update(x)
assert len(nl.get_neighbors(0)[0]) == 12

nl = NeighborList([0.5] * 27, skin=0.01, bothways=True, self_interaction=False)
nl.update(x * (3, 3, 3))
for a in range(27):
    assert len(nl.get_neighbors(a)[0]) == 12
assert not np.any(nl.get_neighbors(13)[1])

c = 0.0058
for NeighborListClass in [PrimitiveNeighborList, NewPrimitiveNeighborList]:
    nl = NeighborListClass([c, c],
                           skin=0.0,
                           sorted=True,
                           self_interaction=False,
                           use_scaled_positions=True)
    nl.update([True, True, True],
              np.eye(3) * 7.56,
              np.array([[0, 0, 0],
                        [0, 0, 0.99875]]))
    n0, d0 = nl.get_neighbors(0)
    n1, d1 = nl.get_neighbors(1)
    # != is xor
    assert (np.all(n0 == [0]) and np.all(d0 == [0, 0, 1])) != \
        (np.all(n1 == [1]) and np.all(d1 == [0, 0, -1]))

# Test empty neighbor list
nl = PrimitiveNeighborList([])
nl.update([True, True, True],
          np.eye(3) * 7.56,
          np.zeros((0, 3)))

# Test hexagonal cell and large cutoff
pbc_c = np.array([True, True, True])
cutoff_a = np.array([8.0, 8.0])
cell_cv = np.array([[0., 3.37316113, 3.37316113],
                    [3.37316113, 0., 3.37316113],
                    [3.37316113, 3.37316113, 0.]])
spos_ac = np.array([[0., 0., 0.],
                    [0.25, 0.25, 0.25]])

nl = PrimitiveNeighborList(cutoff_a, skin=0.0, sorted=True, use_scaled_positions=True)
nl2 = NewPrimitiveNeighborList(cutoff_a, skin=0.0, sorted=True, use_scaled_positions=True)
nl.update(pbc_c, cell_cv, spos_ac)
nl2.update(pbc_c, cell_cv, spos_ac)

a0, offsets0 = nl.get_neighbors(0)
b0 = np.zeros_like(a0)
d0 = np.dot(spos_ac[a0] + offsets0 - spos_ac[0], cell_cv)
a1, offsets1 = nl.get_neighbors(1)
d1 = np.dot(spos_ac[a1] + offsets1 - spos_ac[1], cell_cv)
b1 = np.ones_like(a1)

a = np.concatenate([a0, a1])
b = np.concatenate([b0, b1])
d = np.concatenate([d0, d1])
_a = np.concatenate([a, b])
_b = np.concatenate([b, a])
a = _a
b = _b
d = np.concatenate([d, -d])

a0, offsets0 = nl2.get_neighbors(0)
d0 = np.dot(spos_ac[a0] + offsets0 - spos_ac[0], cell_cv)
b0 = np.zeros_like(a0)
a1, offsets1 = nl2.get_neighbors(1)
d1 = np.dot(spos_ac[a1] + offsets1 - spos_ac[1], cell_cv)
b1 = np.ones_like(a1)

a2 = np.concatenate([a0, a1])
b2 = np.concatenate([b0, b1])
d2 = np.concatenate([d0, d1])
_a2 = np.concatenate([a2, b2])
_b2 = np.concatenate([b2, a2])
a2 = _a2
b2 = _b2
d2 = np.concatenate([d2, -d2])

i = np.argsort(d[:, 0]+d[:, 1]*1e2+d[:, 2]*1e4+a*1e6)
i2 = np.argsort(d2[:, 0]+d2[:, 1]*1e2+d2[:, 2]*1e4+a2*1e6)

assert np.all(a[i] == a2[i2])
assert np.all(b[i] == b2[i2])
assert np.allclose(d[i], d2[i2])



# creates: a1.png, a2.png, a3.png, cnt1.png, cnt2.png, gnr1.png, gnr2.png
from ase.io import write
from ase.build import bulk
from ase.build import nanotube, graphene_nanoribbon

for i, a in enumerate(
    [bulk('Cu', 'fcc', a=3.6),
     bulk('Cu', 'fcc', a=3.6, orthorhombic=True),
     bulk('Cu', 'fcc', a=3.6, cubic=True)]):
    write('a%d.pov' % (i + 1), a,
          run_povray=True)

cnt1 = nanotube(6, 0, length=4, vacuum=2.5)
cnt1.rotate('x', 'z', rotate_cell=True)
cnt2 = nanotube(3, 3, length=6, bond=1.4, symbol='Si', vacuum=2.5)
cnt2.rotate('x', 'z', rotate_cell=True)

for i, a in enumerate([cnt1, cnt2]):
    write('cnt%d.pov' % (i + 1), a,
          run_povray=True)

gnr1 = graphene_nanoribbon(3, 4, type='armchair', saturated=True, vacuum=2.5)
gnr2 = graphene_nanoribbon(2, 6, type='zigzag', saturated=True,
                           C_H=1.1, C_C=1.4, vacuum=3.0,
                           magnetic=True, initial_mag=1.12)

for i, a in enumerate([gnr1, gnr2]):
    write('gnr%d.pov' % (i + 1), a,
          rotation='90x',
          run_povray=True)



import numpy as np

from ase import units
from ase.calculators.calculator import Calculator, all_changes
from ase.calculators.tip3p import rOH, angleHOH, TIP3P

__all__ = ['rOH', 'angleHOH', 'TIP4P', 'sigma0', 'epsilon0']

# Electrostatic constant and parameters:
k_c = units.Hartree * units.Bohr
qH = 0.52
A = 600e3 * units.kcal / units.mol
B = 610 * units.kcal / units.mol
sigma0 = (A / B)**(1 / 6.)
epsilon0 = B**2 / (4 * A)
# http://dx.doi.org/10.1063/1.445869

class TIP4P(TIP3P):
    def __init__(self, rc=7.0, width=1.0):
        """ TIP4P potential for water.

        http://dx.doi.org/10.1063/1.445869

        Requires an atoms object of OHH,OHH, ... sequence
        Correct TIP4P charges and LJ parameters set automatically.

        Virtual interaction sites implemented in the following scheme:
        Original atoms object has no virtual sites.
        When energy/forces are requested:

        * virtual sites added to temporary xatoms object
        * energy / forces calculated
        * forces redistributed from virtual sites to actual atoms object

        This means you do not get into trouble when propagating your system
        with MD while having to skip / account for massless virtual sites.

        This also means that if using for QM/MM MD with GPAW, the EmbedTIP4P
        class must be used.
        """

        TIP3P.__init__(self, rc, width)
        self.atoms_per_mol = 3
        self.sites_per_mol = 4
        self.energy = None
        self.forces = None

    def calculate(self, atoms=None,
                  properties=['energy', 'forces'],
                  system_changes=all_changes):
        Calculator.calculate(self, atoms, properties, system_changes)

        assert (atoms.numbers[::3] == 8).all()
        assert (atoms.numbers[1::3] == 1).all()
        assert (atoms.numbers[2::3] == 1).all()

        xpos = self.add_virtual_sites(atoms.positions)
        xcharges = self.get_virtual_charges(atoms)

        cell = atoms.cell
        pbc = atoms.pbc

        natoms = len(atoms)
        nmol = natoms // 3

        self.energy = 0.0
        self.forces = np.zeros((4 * natoms // 3, 3))

        C = cell.diagonal()
        assert (cell == np.diag(C)).all(), 'not orthorhombic'
        assert ((C >= 2 * self.rc) | ~pbc).all(), 'cutoff too large'

        # Get dx,dy,dz from first atom of each mol to same atom of all other
        # and find min. distance. Everything moves according to this analysis.
        for a in range(nmol - 1):
            D = xpos[(a + 1) * 4::4] - xpos[a * 4]
            shift = np.zeros_like(D)
            for i, periodic in enumerate(pbc):
                if periodic:
                    shift[:, i] = np.rint(D[:, i] / C[i]) * C[i]
            q_v = xcharges[(a + 1) * 4:]

            # Min. img. position list as seen for molecule !a!
            position_list = np.zeros(((nmol - 1 - a) * 4, 3))

            for j in range(4):
                position_list[j::4] += xpos[(a + 1) * 4 + j::4] - shift

            # Make the smooth cutoff:
            pbcRoo = position_list[::4] - xpos[a * 4]
            pbcDoo = np.sum(np.abs(pbcRoo)**2, axis=-1)**(1 / 2)
            x1 = pbcDoo > self.rc - self.width
            x2 = pbcDoo < self.rc
            x12 = np.logical_and(x1, x2)
            y = (pbcDoo[x12] - self.rc + self.width) / self.width
            t = np.zeros(len(pbcDoo))
            t[x2] = 1.0
            t[x12] -= y**2 * (3.0 - 2.0 * y)
            dtdd = np.zeros(len(pbcDoo))
            dtdd[x12] -= 6.0 / self.width * y * (1.0 - y)
            self.energy_and_forces(a, xpos, position_list, q_v, nmol, t, dtdd)

        if self.pcpot:
            e, f = self.pcpot.calculate(xcharges, xpos)
            self.energy += e
            self.forces += f

        f = self.redistribute_forces(self.forces)

        self.results['energy'] = self.energy
        self.results['forces'] = f

    def energy_and_forces(self, a, xpos, position_list, q_v, nmol, t, dtdd):
        """ energy and forces on molecule a from all other molecules.
            cutoff is based on O-O Distance. """

        # LJ part - only O-O interactions
        epsil = np.tile([epsilon0], nmol - 1 - a)
        sigma = np.tile([sigma0], nmol - 1 - a)
        DOO = position_list[::4] - xpos[a * 4]
        d2 = (DOO**2).sum(1)
        d = np.sqrt(d2)
        e_lj = 4 * epsil * (sigma**12 / d**12 - sigma**6 / d**6)
        f_lj = (4 * epsil * (12 * sigma**12 / d**13 -
                             6 * sigma**6 / d**7) * t -
                e_lj * dtdd)[:, np.newaxis] * DOO / d[:, np.newaxis]

        self.forces[a * 4] -= f_lj.sum(0)
        self.forces[(a + 1) * 4::4] += f_lj

        # Electrostatics
        e_elec = 0
        all_cut = np.repeat(t, 4)
        for i in range(4):
            D = position_list - xpos[a * 4 + i]
            d2_all = (D**2).sum(axis=1)
            d_all = np.sqrt(d2_all)
            e = k_c * q_v[i] * q_v / d_all
            e_elec += np.dot(all_cut, e).sum()
            e_f = e.reshape(nmol - a - 1, 4).sum(1)
            F = (e / d_all * all_cut)[:, np.newaxis] * D / d_all[:, np.newaxis]
            FOO = -(e_f * dtdd)[:, np.newaxis] * DOO / d[:, np.newaxis]
            self.forces[(a + 1) * 4 + 0::4] += FOO
            self.forces[a * 4] -= FOO.sum(0)
            self.forces[(a + 1) * 4:] += F
            self.forces[a * 4 + i] -= F.sum(0)

        self.energy += np.dot(e_lj, t) + e_elec

    def add_virtual_sites(self, pos):
        # Order: OHHM,OHHM,...
        # DOI: 10.1002/(SICI)1096-987X(199906)20:8
        b = 0.15
        xatomspos = np.zeros((4 * len(pos) // 3, 3))
        for w in range(0, len(pos), 3):
            r_i = pos[w]  # O pos
            r_j = pos[w + 1]  # H1 pos
            r_k = pos[w + 2]  # H2 pos
            n = (r_j + r_k) / 2 - r_i
            n /= np.linalg.norm(n)
            r_d = r_i + b * n

            x = 4 * w // 3
            xatomspos[x + 0] = r_i
            xatomspos[x + 1] = r_j
            xatomspos[x + 2] = r_k
            xatomspos[x + 3] = r_d

        return xatomspos

    def get_virtual_charges(self, atoms):
        charges = np.empty(len(atoms) * 4 // 3)
        charges[0::4] = 0.00  # O
        charges[1::4] = qH  # H1
        charges[2::4] = qH  # H2
        charges[3::4] = - 2* qH  # X1
        return charges

    def redistribute_forces(self, forces):
        f = forces
        b = 0.15
        a = 0.5
        pos = self.atoms.positions
        for w in range(0, len(pos), 3):
            r_i = pos[w]  # O pos
            r_j = pos[w + 1]  # H1 pos
            r_k = pos[w + 2]  # H2 pos
            r_ij = r_j - r_i
            r_jk = r_k - r_j
            r_d = r_i + b * (r_ij + a * r_jk) / np.linalg.norm(r_ij + a * r_jk)
            r_id = r_d - r_i
            gamma = b / np.linalg.norm(r_ij + a * r_jk)

            x = w * 4 // 3
            Fd = f[x + 3]  # force on M
            F1 = (np.dot(r_id, Fd) / np.dot(r_id, r_id)) * r_id
            Fi = Fd - gamma * (Fd - F1)  # Force from M on O
            Fj = (1 - a) * gamma * (Fd - F1)  # Force from M on H1
            Fk = a * gamma * (Fd - F1)  # Force from M on H2

            f[x] += Fi
            f[x + 1] += Fj
            f[x + 2] += Fk

        # remove virtual sites from force array
        f = np.delete(f, list(range(3, f.shape[0], 4)), axis=0)
        return f



import matplotlib.pyplot as plt
from gpaw import GPAW
from ase.dft.dos import DOS

calc = GPAW('bulk.Ag.gpw')
#energies, weights = calc.get_dos(npts=800, width=0)

dos = DOS(calc, npts=800, width=0)
energies = dos.get_energies()
weights = dos.get_dos()

ax = plt.gca()
ax.plot(energies, weights)
ax.set_xlabel('Energy [eV]')
ax.set_ylabel('DOS [1/eV]')
plt.savefig('dos.png')
plt.show()



import numpy as np
from ase import Atoms
from ase.build import fcc111
from ase.optimize import BFGS
from ase.calculators.emt import EMT as OrigEMT
from ase.neb import NEB

# Global counter of force evaluations:
force_evaluations = [0]


class EMT(OrigEMT):
    def calculate(self, *args, **kwargs):
        force_evaluations[0] += 1
        OrigEMT.calculate(self, *args, **kwargs)


# Build Pt(111) slab with six surface atoms and add oxygen adsorbate
initial = fcc111('Pt', size=(3, 2, 3), orthogonal=True)
initial.center(axis=2, vacuum=10)
oxygen = Atoms('O')
oxygen.translate(initial[7].position + (0., 0., 3.5))
initial.extend(oxygen)

# EMT potential
initial.set_calculator(EMT())

# Optimize initial state
opt = BFGS(initial)
opt.run(fmax=0.03)

# Move oxygen adsorbate to neighboring hollow site
final = initial.copy()
final[18].x += 2.8
final[18].y += 1.8

final.set_calculator(EMT())

opt = BFGS(final)
opt.run(fmax=0.03)

# NEB with seven interior images
images = [initial]
for i in range(7):
    images.append(initial.copy())
images.append(final)

fmax = 0.03  # Same for NEB and optimizer

for i in range(1, len(images)-1):
    calc = EMT()
    images[i].set_calculator(calc)


def run_NEB():
    if method == 'dyn':
        neb = NEB(images, fmax=fmax, dynamic_relaxation=True)
        neb.interpolate()
    elif method == 'dyn_scale':
        neb = NEB(images, fmax=fmax, dynamic_relaxation=True, scale_fmax=6.)
        neb.interpolate()
    else:
        # Default NEB
        neb = NEB(images)
        neb.interpolate()

    # Optimize and check number of calculations.
    # We use a hack with a global counter to count the force evaluations:
    force_evaluations[0] = 0
    opt = BFGS(neb)
    opt.run(fmax=fmax)
    force_calls.append(force_evaluations[0])

    # Get potential energy of transition state.
    Emax.append(np.sort([image.get_potential_energy()
                        for image in images[1:-1]])[-1])


force_calls, Emax = [], []
for method in ['def', 'dyn', 'dyn_scale']:
    run_NEB()

# Check force calculation count for default and dynamic NEB implementations
print('\n# Force calls with default NEB: {}'.format(force_calls[0]))
print('# Force calls with dynamic NEB: {}'.format(force_calls[1]))
print('# Force calls with dynamic and scaled NEB: {}\n'.format(force_calls[2]))
assert force_calls[2] < force_calls[1] < force_calls[0]

# Assert reaction barriers are within 1 meV of default NEB
assert(abs(Emax[1] - Emax[0]) < 1e-3)
assert(abs(Emax[2] - Emax[0]) < 1e-3)



""" An example for using gromacs calculator in ase.
    Atom positions are relaxed.
    A sample call:

   python ./gromacs_example_mm_relax.py his.pdb
"""

from ase.calculators.gromacs import Gromacs

import sys

infile_name = sys.argv[1]

CALC_MM_RELAX = Gromacs(clean=True)
CALC_MM_RELAX.set_own_params_runs(
    'extra_pdb2gmx_parameters', '-ignh')
CALC_MM_RELAX.set_own_params_runs(
    'init_structure', infile_name)
CALC_MM_RELAX.generate_topology_and_g96file()
CALC_MM_RELAX.write_input()
CALC_MM_RELAX.set_own_params_runs(
    'extra_editconf_parameters', '-bt cubic -c -d 0.8')
CALC_MM_RELAX.run_editconf()
CALC_MM_RELAX.set_own_params_runs(
    'extra_genbox_parameters', '-cs spc216.gro')
CALC_MM_RELAX.run_genbox()
CALC_MM_RELAX.generate_gromacs_run_file()
CALC_MM_RELAX.run()



"""
ASE Calculator for interatomic models compatible with the Knowledgebase
of Interatomic Models (KIM) application programming interface (API).
Written by:

Mingjian Wen
Daniel S. Karls
University of Minnesota
"""
import numpy as np

from ase.calculators.calculator import Calculator
from ase.calculators.calculator import compare_atoms

from . import kimpy_wrappers
from . import neighborlist


class KIMModelData(object):
    """Initializes and subsequently stores the KIM API Portable Model
    object, KIM API ComputeArguments object, and the neighbor list
    object used by instances of KIMModelCalculator.  Also stores the
    arrays which are registered in the KIM API and which are used to
    communicate with the model.
    """

    def __init__(self, model_name, ase_neigh, neigh_skin_ratio, debug=False):
        self.model_name = model_name
        self.ase_neigh = ase_neigh
        self.debug = debug

        # Initialize KIM API Portable Model object and ComputeArguments object
        self.init_kim()

        # Set cutoff
        model_influence_dist = self.kim_model.get_influence_distance()
        model_cutoffs, padding_not_require_neigh = (
            self.kim_model.get_neighbor_list_cutoffs_and_hints()
        )

        self.species_map = self.create_species_map()

        # Initialize neighbor list object
        self.init_neigh(
            neigh_skin_ratio,
            model_influence_dist,
            model_cutoffs,
            padding_not_require_neigh,
        )

    def __del__(self):
        self.clean()

    def init_kim(self):
        """Create the KIM API Portable Model object and KIM API ComputeArguments
        object
        """
        if self.kim_initialized:
            return

        self.kim_model = kimpy_wrappers.PortableModel(self.model_name, self.debug)

        # KIM API model object is what actually creates/destroys the ComputeArguments
        # object, so we must pass it as a parameter
        self.compute_args = self.kim_model.compute_arguments_create()

    def init_neigh(
        self,
        neigh_skin_ratio,
        model_influence_dist,
        model_cutoffs,
        padding_not_require_neigh,
    ):
        """Initialize neighbor list, either an ASE-native neighborlist
        or one created using the neighlist module in kimpy
        """
        neigh_list_object_type = (
            neighborlist.ASENeighborList
            if self.ase_neigh
            else neighborlist.KimpyNeighborList
        )
        self.neigh = neigh_list_object_type(
            self.compute_args,
            neigh_skin_ratio,
            model_influence_dist,
            model_cutoffs,
            padding_not_require_neigh,
            self.debug,
        )

    def update_compute_args_pointers(self, energy, forces):
        self.compute_args.update(
            self.num_particles,
            self.species_code,
            self.particle_contributing,
            self.coords,
            energy,
            forces,
        )

    def create_species_map(self):
        """Get all the supported species of the KIM model and the
        corresponding integer codes used by the model

        Returns
        -------
        species_map : dict
            key : str
                chemical symbols (e.g. "Ar")
            value : int
                species integer code (e.g. 1)
        """
        supported_species, codes = self.get_model_supported_species_and_codes()
        species_map = dict()
        for i, spec in enumerate(supported_species):
            species_map[spec] = codes[i]
            if self.debug:
                print(
                    "Species {} is supported and its code is: {}".format(spec, codes[i])
                )

        return species_map

    def clean_neigh(self):
        """If the neighbor list method being used is the one in the
        kimpy neighlist module, deallocate its memory
        """
        if self.neigh_initialized:
            self.neigh.clean()
            del self.neigh

    def clean_kim(self):
        """Deallocate the memory allocated to the KIM API Portable Model object
        and KIM API ComputeArguments object
        """
        if self.kim_initialized:
            self.kim_model.compute_arguments_destroy(self.compute_args)
            self.kim_model.destroy()
            del self.kim_model

    def clean(self):
        """Deallocate the KIM API Portable Model object, KIM API ComputeArguments
        object, and, if applicable, the neighbor list object
        """
        self.clean_neigh()
        self.clean_kim()

    @property
    def padding_image_of(self):
        return self.neigh.padding_image_of

    @property
    def num_particles(self):
        return self.neigh.num_particles

    @property
    def coords(self):
        return self.neigh.coords

    @property
    def particle_contributing(self):
        return self.neigh.particle_contributing

    @property
    def species_code(self):
        return self.neigh.species_code

    @property
    def kim_initialized(self):
        return hasattr(self, "kim_model")

    @property
    def neigh_initialized(self):
        return hasattr(self, "neigh")

    @property
    def get_model_supported_species_and_codes(self):
        return self.kim_model.get_model_supported_species_and_codes


class KIMModelCalculator(Calculator):
    """Calculator that works with KIM Portable Models (PMs).

    Calculator that carries out direct communication between ASE and a
    KIM Portable Model (PM) through the kimpy library (which provides a
    set of python bindings to the KIM API).

    Parameters
    ----------
    model_name : str
      The unique identifier assigned to the interatomic model (for
      details, see https://openkim.org/doc/schema/kim-ids)

    ase_neigh : bool, optional
      False (default): Use kimpy's neighbor list library

      True: Use ASE's internal neighbor list mechanism (usually slower
      than the kimpy neighlist library)

    neigh_skin_ratio : float, optional
      Used to determine the neighbor list cutoff distance, r_neigh,
      through the relation r_neigh = (1 + neigh_skin_ratio) * rcut,
      where rcut is the model's influence distance. (Default: 0.2)

    release_GIL : bool, optional
      Whether to release python GIL.  Releasing the GIL allows a KIM
      model to run with multiple concurrent threads. (Default: False)

    debug : bool, optional
      If True, detailed information is printed to stdout. (Default:
      False)
    """

    implemented_properties = ["energy", "forces", "stress"]

    def __init__(
        self,
        model_name,
        ase_neigh=False,
        neigh_skin_ratio=0.2,
        release_GIL=False,
        debug=False,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        self.model_name = model_name
        self.release_GIL = release_GIL
        self.debug = debug

        if neigh_skin_ratio < 0:
            raise ValueError('Argument "neigh_skin_ratio" must be non-negative')

        # Model output
        self.energy = None
        self.forces = None

        # Create KIMModelData object. This will take care of creating and storing the KIM
        # API Portable Model object, KIM API ComputeArguments object, and the neighbor
        # list object that our calculator needs
        self.kimmodeldata = KIMModelData(
            self.model_name, ase_neigh, neigh_skin_ratio, self.debug
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, value, traceback):
        # Explicitly deallocate all three objects held by the KIMModelData
        # instance referenced by our calculator
        self.kimmodeldata.clean()

    def __repr__(self):
        return "KIMModelCalculator(model_name={})".format(self.model_name)

    def calculate(
        self,
        atoms=None,
        properties=["energy", "forces", "stress"],
        system_changes=["positions", "numbers", "cell", "pbc"],
    ):
        """
        Inherited method from the ase Calculator class that is called by
        get_property()

        Parameters
        ----------
        atoms : Atoms
            Atoms object whose properties are desired

        properties : list of str
            List of what needs to be calculated.  Can be any combination
            of 'energy', 'forces' and 'stress'.

        system_changes : list of str
            List of what has changed since last calculation.  Can be any
            combination of these six: 'positions', 'numbers', 'cell',
            and 'pbc'.
        """

        Calculator.calculate(self, atoms, properties, system_changes)

        # Update KIM API input data and neighbor list, if necessary
        if system_changes:
            if self.need_neigh_update(atoms, system_changes):
                self.update_neigh(atoms, self.species_map)
                self.energy = np.array([0.0], dtype=np.double)
                self.forces = np.zeros([self.num_particles[0], 3], dtype=np.double)
                self.update_compute_args_pointers(self.energy, self.forces)
            else:
                self.update_kim_coords(atoms)

            self.kim_model.compute(self.compute_args, self.release_GIL)

        energy = self.energy[0]
        forces = self.assemble_padding_forces()

        try:
            volume = atoms.get_volume()
            stress = self.compute_virial_stress(self.forces, self.coords, volume)
        except ValueError:  # Volume cannot be computed
            stress = None

        # Quantities passed back to ASE
        self.results["energy"] = energy
        self.results["free_energy"] = energy
        self.results["forces"] = forces
        self.results["stress"] = stress

    def check_state(self, atoms, tol=1e-15):
        return compare_atoms(self.atoms, atoms, excluded_properties={'initial_charges',
            'initial_magmoms'})

    def assemble_padding_forces(self):
        """
        Assemble forces on padding atoms back to contributing atoms.

        Parameters
        ----------
        forces : 2D array of doubles
            Forces on both contributing and padding atoms

        num_contrib:  int
            Number of contributing atoms

        padding_image_of : 1D array of int
            Atom number, of which the padding atom is an image


        Returns
        -------
            Total forces on contributing atoms.
        """

        total_forces = np.array(self.forces[: self.num_contributing_particles])

        if self.padding_image_of.size != 0:
            pad_forces = self.forces[self.num_contributing_particles :]
            for f, org_index in zip(pad_forces, self.padding_image_of):
                total_forces[org_index] += f

        return total_forces

    @staticmethod
    def compute_virial_stress(forces, coords, volume):
        """Compute the virial stress in Voigt notation.

        Parameters
        ----------
        forces : 2D array
            Partial forces on all atoms (padding included)

        coords : 2D array
            Coordinates of all atoms (padding included)

        volume : float
            Volume of cell

        Returns
        -------
        stress : 1D array
            stress in Voigt order (xx, yy, zz, yz, xz, xy)
        """
        stress = np.zeros(6)
        stress[0] = -np.dot(forces[:, 0], coords[:, 0]) / volume
        stress[1] = -np.dot(forces[:, 1], coords[:, 1]) / volume
        stress[2] = -np.dot(forces[:, 2], coords[:, 2]) / volume
        stress[3] = -np.dot(forces[:, 1], coords[:, 2]) / volume
        stress[4] = -np.dot(forces[:, 0], coords[:, 2]) / volume
        stress[5] = -np.dot(forces[:, 0], coords[:, 1]) / volume

        return stress

    def get_model_supported_species_and_codes(self):
        return self.kimmodeldata.get_model_supported_species_and_codes

    @property
    def update_compute_args_pointers(self):
        return self.kimmodeldata.update_compute_args_pointers

    @property
    def kim_model(self):
        return self.kimmodeldata.kim_model

    @property
    def compute_args(self):
        return self.kimmodeldata.compute_args

    @property
    def num_particles(self):
        return self.kimmodeldata.num_particles

    @property
    def coords(self):
        return self.kimmodeldata.coords

    @property
    def padding_image_of(self):
        return self.kimmodeldata.padding_image_of

    @property
    def species_map(self):
        return self.kimmodeldata.species_map

    @property
    def neigh(self):
        return self.kimmodeldata.neigh

    @property
    def num_contributing_particles(self):
        return self.neigh.num_contributing_particles

    @property
    def update_kim_coords(self):
        return self.neigh.update_kim_coords

    @property
    def need_neigh_update(self):
        return self.neigh.need_neigh_update

    @property
    def update_neigh(self):
        return self.neigh.update



from ase import Atoms
import numpy as np

atoms = Atoms(['O', 'H', 'H'], positions=[[0., 0., 0.119262],
                                          [0., 0.763239, -0.477047],
                                          [0., -0.763239, -0.477047]])

# Angle no pbc
assert abs(atoms.get_angle(1, 0, 2) - 104) < 1e-3

atoms.set_cell([2, 2, 2])

# Across different pbcs
atoms.set_pbc([True, False, True])
atoms.wrap()
assert abs(atoms.get_angle(1, 0, 2, mic=True) - 104) < 1e-3

# Across all True pbc
atoms.set_pbc(True)
atoms.wrap()
assert abs(atoms.get_angle(1, 0, 2, mic=True) - 104) < 1e-3

# Change Angle
old = atoms.get_angle(1, 0, 2, mic=False)
atoms.set_angle(1, 0, 2, -10, indices=[2], add=True)
new = atoms.get_angle(1, 0, 2, mic=False)
diff = old - new - 10
assert abs(diff) < 10e-3

#don't actually change angle using indices
old = atoms.get_angle(1, 0, 2, mic=False)
atoms.set_angle(1, 0, 2, -10, indices=[2, 1], add=True)
new = atoms.get_angle(1, 0, 2, mic=False)
diff = old - new
assert abs(diff) < 10e-3

# Simple tetrahedron
tetra_pos = np.array([[0, 0, 0], [1, 0, 0], [.5, np.sqrt(3) * .5, 0],
                      [.5, np.sqrt(1/3.) * .5, np.sqrt(2/3.)]])
atoms = Atoms(['H', 'H', 'H', 'H'],
              positions=tetra_pos - np.array([.2, 0, 0]))
angle = 70.5287793655
assert abs(atoms.get_dihedral(0, 1, 2, 3) - angle) < 1e-3

atoms.set_cell([3, 3, 3])
atoms.set_pbc(True)
atoms.wrap()
assert abs(atoms.get_dihedral(0, 1, 2, 3, mic=True) - angle) < 1e-3





import numpy as np


class DOS:
    def __init__(self, energy, weights, info=None, sampling={'type': 'raw'}):
        """
        Docstring here
        """
        self.energy = np.asarray(energy)
        self.weights = np.asarray(weights)
        self.sampling = sampling

        # Energy format: [e1, e2, ...]
        if self.energy.ndim != 1:
            msg = ('Incorrect Energy dimensionality. '
                   'Expected 1 got {}'.format(
                       self.energy.ndim))
            raise ValueError(msg)

        # Weights format: [[w1, w2, ...], [w1, w2, ..], ...]
        if self.weights.ndim != 2:
            msg = ('Incorrect weight dimensionality. '
                   'Expected 2, got {}'.format(
                       self.weights.ndim))
            raise ValueError(msg)

        # Check weight shape matches energy
        if self.weights.shape[1] != self.energy.shape[0]:
            msg = ('Weight dimensionality does not match energy.'
                   ' Expected {}, got {}'.format(self.energy.shape[0],
                                                 self.weights.shape[1]))
            raise ValueError(msg)

        # One entry for info for each weight
        if info is None:
            info = [{} for _ in self.weights]
        else:
            if len(info) != len(weights):
                msg = ('Incorrect number of entries in '
                       'info. Expected {}, got {}'.format(
                           len(self.weights), len(info)))
                raise ValueError(msg)
        self.info = np.asarray(info)  # Make info np array for slicing purposes

    def delta(self, x, x0, width, smearing='Gauss'):
        """Return a delta-function centered at 'x0'."""
        if smearing.lower() == 'gauss':
            x1 = -((x - x0) / width)**2
            return np.exp(x1) / (np.sqrt(np.pi) * width)
        else:
            msg = 'Requested smearing type not recognized. Got {}'.format(
                smearing)
            raise ValueError(msg)

    def smear(self, energy_grid, width=0.1, smearing='Gauss'):
        """Add Gaussian smearing, to all weights onto an energy grid.
        Disabled for width=0.0"""
        if width == 0.0:
            msg = 'Cannot add 0 width smearing'
            raise ValueError(msg)

        en0 = self.energy[:, np.newaxis]  # Add axis to use NumPy broadcasting
        weights_grid = np.dot(self.weights,
                              self.delta(energy_grid, en0, width,
                                         smearing=smearing))

        return weights_grid

    def sample(self, grid, width=0.1, smearing='Gauss', gridtype='general'):
        """Sample this DOS on a grid, returning result as a new DOS."""

        npts = len(grid)
        sampling = {'width': width,
                    'smearing': smearing,
                    'npts': npts,
                    'type': gridtype}

        weights_grid = self.smear(grid, width=width, smearing=smearing)

        dos_new = DOS(grid, weights_grid,
                      info=self.info, sampling=sampling)
        return dos_new

    def sample_grid(self, spacing=None, npts=None, width=0.1,
                    window=None, smearing='Gauss'):
        """Sample this DOS on a uniform grid, returning result as a new DOS."""

        if window is None:
            emin, emax = None, None
        else:
            emin, emax = window

        if emin is None:
            emin = self.energy.min()
        if emax is None:
            emax = self.energy.max()
        emin -= 5 * width
        emax += 5 * width

        grid_uniform = DOS._make_uniform_grid(emin, emax, spacing=spacing,
                                              npts=npts, width=width)

        return self.sample(grid_uniform, width=width,
                           smearing=smearing, gridtype='uniform')

    @staticmethod
    def sample_many(doslist, grid, width=0.1, smearing='Gauss',
                    gridtype='general'):
        """Take list of DOS objects, and combine into 1, with same grid."""

        # Count the total number of weights
        n_weights = sum(len(dos.weights) for dos in doslist)

        npts = len(grid)

        weight_grid = np.zeros((n_weights, npts))
        info_new = []
        # Do sampling
        ii = 0
        for dos in doslist:
            dos_sample = dos.sample(grid, width=width,
                                    smearing=smearing)
            info_new.extend(dos_sample.info)
            for w_i in dos_sample.weights:
                weight_grid[ii] = w_i
                ii += 1
        sampling = {'smearing': smearing,
                    'width': width,
                    'npts': npts,
                    'type': gridtype}
        return DOS(energy=grid, weights=weight_grid, info=info_new,
                   sampling=sampling)

    @staticmethod
    def sample_many_grid(doslist, window=None, spacing=None,
                         npts=None, width=0.1, smearing='Gauss'):
        """Combine list of DOS objects onto uniform grid.

        Takes the lowest and highest energies as grid range, if
        no window is specified."""
        dosen = [dos.energy for dos in doslist]
        # Parse window
        if window is None:
            emin, emax = None, None
        else:
            emin, emax = window
        if emin is None:
            emin = np.min(dosen)
        if emax is None:
            emax = np.max(dosen)
        # Add a little extra to avoid stopping midpeak
        emin -= 5 * width
        emax += 5 * width

        grid_uniform = DOS._make_uniform_grid(emin, emax, spacing=spacing,
                                              npts=npts, width=width)

        return DOS.sample_many(doslist, grid_uniform, width=width,
                               smearing=smearing, gridtype='uniform')

    @staticmethod
    def join(doslist, atol=1e-08):
        """Join a list of DOS objects into one, without applying sampling.

        Requires all energies to be identical"""

        # Test if energies are the same
        eneq = all(np.allclose(doslist[0].energy, dos.energy, atol=atol)
                   for dos in doslist)
        if not eneq:
            msg = 'Energies must the the same in all DOS objects.'
            raise ValueError(msg)

        energy = doslist[0].energy     # Just use the first energy
        weights = []
        info = []
        for dos in doslist:
            for info_i, w_i in zip(dos.info, dos.weights):
                weights.append(w_i)
                info.append(info_i)

        return DOS(energy, weights, info=info)

    @staticmethod
    def _make_uniform_grid(emin, emax, spacing=None, npts=None, width=0.1):
        if spacing and npts:
            msg = ('spacing and npts cannot both be defined'
                   ' at the same time.')
            raise ValueError(msg)
        if not spacing and not npts:
            # Default behavior
            spacing = 0.2 * width
        # Now either spacing or npts is defined
        if npts:
            grid_uniform = np.linspace(emin, emax, npts)
        else:
            grid_uniform = np.arange(emin, emax, spacing)
        return grid_uniform

    def plot(self,
             # We need to grab init keywords
             ax=None,
             emin=None, emax=None,
             ymin=None, ymax=None, ylabel=None,
             *plotargs, **plotkwargs):

        pdp = DOSPlot(self, ax=None,
                      emin=None, emax=None,
                      ymin=None, ymax=None, ylabel=None)
        return pdp.plot(*plotargs, **plotkwargs)

    def sum(self):
        """Return the sum of all weights in this DOS as a new DOS."""
        weights_sum = self.weights.sum(0)[np.newaxis]

        # Find shared (key, value) pairs
        # dict(set.intersection(*(set(d.items()) for d in info)))
        all_kv = []
        for d in self.info:
            kv_pairs = set()
            for key, value in d.items():
                try:
                    kv_pairs.add((key, value))
                except TypeError:
                    # Unhashable type, skip it
                    pass
            all_kv.append(kv_pairs)
        if all_kv:
            info_new = [dict(set.intersection(*all_kv))]
        else:
            # We didn't find any shared (key, value) pairs
            # This prevents set.intersection from blowing up
            info_new = None

        return DOS(energy=self.energy, weights=weights_sum,
                   info=info_new, sampling=self.sampling)

    def pick(self, **kwargs):
        """Pick key/value pairs using logical AND
        i.e., all conditions from kwargs must be met"""
        idx = [i for i, d in enumerate(self.info)
               if all(d.get(key) == value
                      for key, value in kwargs.items())]

        return self[idx]

    def split(self, key):
        """Find all unique instances of key in info"""
        unique = np.unique([info.get(key) for info in self.info
                            if info.get(key, None) is not None])

        dos_lst = []
        for value in unique:
            # Use **{key: value} instead of key=value,
            # as key=value will litterally look up "key" in info.
            dos_lst.append(self.pick(**{key: value}))
        return dos_lst

    def __getitem__(self, i):
        if isinstance(i, int):
            n_weights = len(self.weights)
            if i < -n_weights or i >= n_weights:
                raise IndexError('Index out of range.')

        indices = np.arange(len(self.weights))[i]
        if len(indices.shape) == 0:
            indices = indices[np.newaxis]

        return DOS(energy=self.energy,
                   weights=self.weights[indices],
                   info=self.info[indices],
                   sampling=self.sampling)


class DOSPlot:
    def __init__(self, dos, ax=None,
                 emin=None, emax=None,
                 ymin=None, ymax=None, ylabel=None):
        self.dos = dos
        self.ax = ax
        if self.ax is None:
            self.ax = self.prepare_plot(ax, emin, emax,
                                        ymin=ymin, ymax=ymax,
                                        ylabel=ylabel)

    def plot(self, filename=None, show=None, colors=None,
             labels=None, show_legend=True, loc='best', **plotkwargs):

        ax = self.ax

        for ii, w_i in enumerate(self.dos.weights):
            # We can add smater labeling later
            kwargs = {}
            if colors is not None:
                kwargs['color'] = colors[ii]

            # We could possibly have some better label logic here
            if labels is not None:
                kwargs['label'] = labels[ii]
            else:
                kwargs['label'] = self.dos.info[ii]
            kwargs.update(plotkwargs)
            ax.plot(self.dos.energy, w_i,
                    **kwargs)

        self.finish_plot(filename, show, show_legend, loc)

        return ax

    def prepare_plot(self, ax=None, emin=None, emax=None,
                     ymin=None, ymax=None,
                     ylabel=None, xlabel=None):
        import matplotlib.pyplot as plt
        if ax is None:
            ax = plt.figure().add_subplot(111)

        ylabel = ylabel if ylabel is not None else 'DOS'
        xlabel = xlabel if xlabel is not None else 'Energy [eV]'
        ax.axis(xmin=emin, xmax=emax, ymin=ymin, ymax=ymax)
        ax.set_ylabel(ylabel)
        self.ax = ax
        return ax

    def finish_plot(self, filename, show, show_legend, loc):
        import matplotlib.pyplot as plt

        if show_legend:
            leg = plt.legend(loc=loc)
            leg.get_frame().set_alpha(1)

        if filename:
            plt.savefig(filename)

        if show is None:
            show = not filename

        if show:
            plt.show()



import os
import shutil
import importlib
from ase.calculators.calculator import names

builtins = {'eam', 'emt', 'ff', 'lj', 'morse', 'tip3p', 'tip4p'}

required_envvars = {'abinit': ['ABINIT_PP_PATH'],
                    'elk': ['ELK_SPECIES_PATH'],
                    'openmx': ['OPENMX_DFT_DATA_PATH']}

default_executables = {'abinit': ['abinit'],
                       'cp2k': ['cp2k_shell', 'cp2k_shell.psmp',
                                'cp2k_shell.popt', 'cp2k_shell.ssmp',
                                'cp2k_shell.sopt'],
                       'dftb': ['dftb+'],
                       'elk': ['elk', 'elk-lapw'],
                       'espresso': ['pw.x'],
                       'gromacs': ['gmx', 'gmx_d', 'gmx_mpi', 'gmx_mpi_d'],
                       'lammpsrun': ['lammps', 'lmp', 'lmp_mpi', 'lmp_serial'],
                       'mopac': ['mopac', 'run_mopac7'], # run_mopac7: debian
                       'nwchem': ['nwchem'],
                       'octopus': ['octopus'],
                       'openmx': ['openmx'],
                       'psi4': ['psi4'],
                       'siesta': ['siesta'],
                       }

python_modules = {'gpaw': 'gpaw',
                  'asap': 'asap3',
                  'lammpslib': 'lammps'}

def get_executable_env_var(name):
    return 'ASE_{}_COMMAND'.format(name.upper())


def detect(name):
    assert name in names
    d = {'name': name}

    if name in builtins:
        d['type'] = 'builtin'
        return d

    if name in python_modules:
        loader = importlib.find_loader(python_modules[name])
        if loader is not None:
            d['type'] = 'python'
            d['module'] = python_modules[name]
            d['path'] = loader.get_filename()
            return d

    envvar = get_executable_env_var(name)
    if envvar in os.environ:
        d['command'] = os.environ[envvar]
        d['envvar'] = envvar
        d['type'] = 'environment'
        return d

    if name in default_executables:
        commands = default_executables[name]
        for command in commands:
            fullpath = shutil.which(command)
            if fullpath:
                d['command'] = command
                d['fullpath'] = fullpath
                d['type'] = 'which'
                return d


def detect_calculators():
    configs = {}
    for name in names:
        result = detect(name)
        if result:
            configs[name] = result
    return configs


def format_configs(configs):
    messages = []
    for name in names:
        config = configs.get(name)

        if config is None:
            state = 'no'
        else:
            type = config['type']
            if type == 'builtin':
                state = 'yes, builtin: module ase.calculators.{name}'
            elif type == 'python':
                state = 'yes, python: {module} ▶ {path}'
            elif type == 'which':
                state = 'yes, shell command: {command} ▶ {fullpath}'
            else:
                state = 'yes, environment: ${envvar} ▶ {command}'

            state = state.format(**config)

        messages.append('{:<10s} {}'.format(name, state))
    return messages



from ase.calculators.emt import EMT
from ase.db import connect
from ase.build import fcc111, add_adsorbate
from ase.constraints import FixAtoms
from ase.optimize import BFGS

db1 = connect('bulk.db')
db2 = connect('ads.db')


def run(symb, a, n, ads):
    atoms = fcc111(symb, (1, 1, n), a=a)
    add_adsorbate(atoms, ads, height=1.0, position='fcc')

    # Constrain all atoms except the adsorbate:
    fixed = list(range(len(atoms) - 1))
    atoms.constraints = [FixAtoms(indices=fixed)]

    atoms.calc = EMT()
    opt = BFGS(atoms, logfile=None)
    opt.run(fmax=0.01)
    return atoms


for row in db1.select():
    a = row.cell[0, 1] * 2
    symb = row.symbols[0]
    for n in [1, 2, 3]:
        for ads in 'CNO':
            atoms = run(symb, a, n, ads)
            db2.write(atoms, layers=n, surf=symb, ads=ads)



#!/usr/bin/env python3

import os
from numpy.testing import assert_allclose

from ase.build import molecule
from ase.calculators.psi4 import Psi4


def main():
    atoms = molecule('H2O')
    atoms.rotate(30, 'x')

    calc = Psi4(basis='3-21G')
    atoms.set_calculator(calc)

    # Calculate forces ahead of time, compare against finite difference after
    # checking the psi4-calc.dat file
    atoms.get_forces()
    assert_allclose(atoms.get_potential_energy(), -2056.785854116688,
                    rtol=1e-4, atol=1e-4)

    # Test the reader
    calc2 = Psi4()
    calc2.read('psi4-calc')
    assert_allclose(calc2.results['energy'], atoms.get_potential_energy(),
                    rtol=1e-4, atol=1e-4)
    assert_allclose(calc2.results['forces'], atoms.get_forces(),
                    rtol=1e-4, atol=1e-4)

    # Compare analytical vs numerical forces
    assert_allclose(atoms.get_forces(), calc.calculate_numerical_forces(atoms),
                    rtol=1e-4, atol=1e-4)

    os.remove('psi4-calc.dat')
    # Unfortunately, we can't currently remove timer.dat, because Psi4
    # creates the file after this script exits. Not even atexit works.
    # os.remove('timer.dat')


main()



"""This module contains functions to read from QBox output files"""

from ase import Atom, Atoms
from ase.calculators.singlepoint import SinglePointCalculator
from ase.utils import basestring

import re
import xml.etree.ElementTree as ET


# Compile regexs for fixing XML
re_find_bad_xml = re.compile(r'<(/?)([A-z]+) expectation ([a-z]+)')


def read_qbox(f, index=-1):
    """Read data from QBox output file

    Inputs:
        f - str or fileobj, path to file or file object to read from
        index - int or slice, which frames to return

    Returns:
        list of Atoms or atoms, requested frame(s)
    """

    if isinstance(f, basestring):
        f = open(f, 'r')

    # Check whether this is a QB@all output
    version = None
    for line in f:
        if '<release>' in line:
            version = ET.fromstring(line)
            break
    if version is None:
        raise Exception('Parse Error: Version not found')
    is_qball = 'qb@LL' in version.text or 'qball' in version.text

    # Load in atomic species
    species = dict()
    if is_qball:
        # Read all of the lines between release and the first call to `run`
        species_data = []
        for line in f:
            if '<run' in line:
                break
            species_data.append(line)
        species_data = '\n'.join(species_data)

        # Read out the species information with regular expressions
        symbols = re.findall('symbol_ = ([A-Z][a-z]?)', species_data)
        masses = re.findall('mass_ = ([0-9.]+)', species_data)
        names = re.findall('name_ = ([a-z]+)', species_data)
        numbers = re.findall('atomic_number_ = ([0-9]+)', species_data)

        # Compile them into a dictionary
        for name, symbol, mass, number in zip(names, symbols, masses, numbers):
            spec_data = dict(
                symbol=symbol,
                mass=float(mass),
                number=float(number)
            )
            species[name] = spec_data
    else:
        # Find all species
        species_blocks = _find_blocks(f, 'species', '<cmd>run')

        for spec in species_blocks:
            name = spec.get('name')
            spec_data = dict(
                symbol=spec.find('symbol').text,
                mass=float(spec.find('mass').text),
                number=int(spec.find('atomic_number').text))
            species[name] = spec_data

    # Find all of the frames
    frames = _find_blocks(f, 'iteration', None)

    # If index is an int, return one frame
    if isinstance(index, int):
        return _parse_frame(frames[index], species)
    else:
        return [_parse_frame(frame, species) for frame in frames[index]]


def _find_blocks(fp, tag, stopwords='[qbox]'):
    """Find and parse a certain block of the file.

    Reads a file sequentially and stops when it either encounters the end of the file, or until the it encounters a line
    that contains a user-defined string *after it has already found at least one desired block*. Use the stopwords
    ``[qbox]`` to read until the next command is issued.

    Groups the text between the first line that contains <tag> and the next line that contains </tag>, inclusively. The
    function then parses the XML and returns the Element object.

    Inputs:
        fp - file-like object, file to be read from
        tag - str, tag to search for (e.g., 'iteration'). `None` if you want to read until the end of the file
        stopwords - str, halt parsing if a line containing this string is encountered

    Returns:
        list of xml.ElementTree, parsed XML blocks found by this class
    """

    start_tag = '<%s'%tag
    end_tag = '</%s>'%tag

    blocks = []  # Stores all blocks
    cur_block = []  # Block being filled
    in_block = False  # Whether we are currently parsing
    for line in fp:

        # Check if the block has started
        if start_tag in line:
            if in_block:
                raise Exception('Parsing failed: Encountered nested block')
            else:
                in_block = True

        # Add data to block
        if in_block:
            cur_block.append(line)

        # Check for stopping conditions
        if stopwords is not None:
            if stopwords in line and len(blocks) > 0:
                break

        if end_tag in line:
            if in_block:
                blocks.append(cur_block)
                cur_block = []
                in_block = False
            else:
                raise Exception('Parsing failed: End tag found before start tag')

    # Join strings in a block into a single string
    blocks = [''.join(b) for b in blocks]

    # Ensure XML compatibility. There are two specific tags in QBall that are not
    #  valid XML, so we need to run a
    blocks = [re_find_bad_xml.sub(r'<\1\2_expectation_\3', b) for b in blocks]

    # Parse the blocks
    return [ET.fromstring(b) for b in blocks]


def _parse_frame(tree, species):
    """Parse a certain frame from QBOX output

    Inputs:
        tree - ElementTree, <iteration> block from output file
        species - dict, data about species. Key is name of atom type,
            value is data about that type
    Return:
        Atoms object describing this iteration"""

    # Load in data about the system
    energy = float(tree.find("etotal").text)

    # Load in data about the cell
    unitcell = tree.find('atomset').find('unit_cell')
    cell = []
    for d in ['a', 'b', 'c']:
        cell.append([float(x) for x in unitcell.get(d).split()])

    stress_tree = tree.find('stress_tensor')
    if stress_tree is None:
        stresses = None
    else:
        stresses = [float(stress_tree.find('sigma_%s' % x).text)
                    for x in ['xx', 'yy', 'zz', 'yz', 'xz', 'xy']]

    # Create the Atoms object
    atoms = Atoms(pbc=True, cell=cell)

    # Load in the atom information
    forces = []
    for atom in tree.find('atomset').findall('atom'):
        # Load data about the atom type
        spec = atom.get('species')
        symbol = species[spec]['symbol']
        mass = species[spec]['mass']

        # Get data about position / velocity / force
        pos = [float(x) for x in atom.find('position').text.split()]
        force = [float(x) for x in atom.find('force').text.split()]
        momentum = [float(x) * mass
                    for x in atom.find('velocity').text.split()]

        # Create the objects
        atom = Atom(symbol=symbol, mass=mass, position=pos, momentum=momentum)
        atoms += atom
        forces.append(force)

    # Create the calculator object that holds energy/forces
    calc = SinglePointCalculator(atoms,
                                 energy=energy, forces=forces, stress=stresses)
    atoms.set_calculator(calc)

    return atoms



# flake8: noqa
import numpy as np
import xml.etree.ElementTree as ET
from xml.dom import minidom

from ase.utils import basestring
from ase.io.xsd import SetChild, _write_xsd_html
from ase import Atoms

_image_header = ' '*74 + '0.0000\n!DATE     Jan 01 00:00:00 2000\n'
_image_footer = 'end\nend\n'

def _get_atom_str(an,xyz):
    s = '{:<5}'.format(an)
    s += '{:>15.9f}{:>15.9f}{:>15.9f}'.format(xyz[0],xyz[1],xyz[2])
    s += ' XXXX 1      xx      '
    s += '{:<2}'.format(an)
    s += '  0.000\n'
    return s

def write_xtd(filename, images, connectivity=None, moviespeed = 10):
    """Takes Atoms object, and write materials studio file
    atoms: Atoms object
    filename: path of the output file
    moviespeed: speed of animation. between 0 and 10

    note: material studio file cannot use a partial periodic system. If partial
    perodic system was inputted, full periodicity was assumed.
    """
    if moviespeed < 0 or moviespeed > 10:
        raise ValueError('moviespeed only between 0 and 10 allowed')

    if hasattr(images, 'get_positions'):
        images = [images]
        
    XSD,ATR = _write_xsd_html(images,connectivity)
    ATR.attrib['NumChildren'] = '2'
    natoms = len(images[0])
    
    bonds = list()
    if connectivity is not None:
        for i in range(connectivity.shape[0]):
            for j in range(i+1,connectivity.shape[0]):
                if connectivity[i,j]:
                    bonds.append([i,j])

    # non-periodic system
    s = '!BIOSYM archive 3\n'
    if not images[0].pbc.all():
        # Write trajectory
        SetChild(ATR,'Trajectory',{'ID':str(natoms+3+len(bonds)),'Increment':'-1','End':str(len(images)),
            'Type':'arc','Speed':str(moviespeed),'FrameClassType':'Atom'})
        
        # write frame information file
        s += 'PBC=OFF\n'
        for image in images:
            s += _image_header
            s +='\n'
            an = image.get_chemical_symbols()
            xyz = image.get_positions()
            for i in range(natoms):
                s += _get_atom_str(an[i],xyz[i,:])
            s += _image_footer

    # periodic system
    else:
        SetChild(ATR,'Trajectory',{'ID':str(natoms+9+len(bonds)),'Increment':'-1','End':str(len(images)),
            'Type':'arc','Speed':str(moviespeed),'FrameClassType':'Atom'})
            
        # write frame information file
        s += 'PBC=ON\n'
        for image in images:
            s += _image_header
            s += 'PBC'
            vec = image.cell.lengths()
            s+='{:>10.4f}{:>10.4f}{:>10.4f}'.format(vec[0],vec[1],vec[2])
            angles = image.cell.angles() 
            s+='{:>10.4f}{:>10.4f}{:>10.4f}'.format(angles[0],angles[1],angles[2])
            s+='\n'
            an = image.get_chemical_symbols()

            angrad = np.deg2rad(angles)
            cell = np.zeros((3,3))
            cell[0,:] = [vec[0],0 ,0]
            cell[1,:] = np.array([np.cos(angrad[2]), np.sin(angrad[2]), 0]) * vec[1]
            cell[2,0] = vec[2]*np.cos(angrad[1])
            cell[2,1] = (vec[1]*vec[2]*np.cos(angrad[0])-cell[1,0]*cell[2,0])/cell[1,1]
            cell[2,2] = np.sqrt(vec[2]**2-cell[2,0]**2-cell[2,1]**2)
            xyz = np.dot(image.get_scaled_positions(),cell)
            for i in range(natoms):
                s += _get_atom_str(an[i],xyz[i,:])
            s += _image_footer

    # print arc file
    if isinstance(filename,str):
        farcname = filename[:-3] + 'arc'
    else:
        farcname = filename.name[:-3] + 'arc'
    farc = open(farcname, 'w')
    farc.write(s)
    farc.close()

    # check if file is an object or not.
    if isinstance(filename, basestring):
        f = open(filename, 'w')
    else:  # Assume it's a 'file-like object'
        f = filename

    # Return a pretty-printed XML string for the Element.
    rough_string = ET.tostring(XSD, 'utf-8')
    reparsed = minidom.parseString(rough_string)
    Document = reparsed.toprettyxml(indent='\t')

    # write
    f.write(Document)
    f.close()
    
def read_xtd(filename, index=-1):
    """Import xtd file (Materials Studio)
    
    Xtd files always come with arc file, and arc file
    contains all the relevant information to make atoms
    so only Arc file needs to be read
    """
    if isinstance(filename,str):
        f = filename[:-3] + 'arc'
    else:
        f = filename.name[:-3] + 'arc'
    f = open(f,'r')
    images = list()

    # the first line is comment
    f.readline()
    pbc = 'ON' in f.readline()
    l = f.readline()
    while l != '':
        if '!' not in l: # flag for the start of an image
            l = f.readline()
            continue
        if pbc:
            l = f.readline()
            cell = [float(d) for d in l.split()[1:]]
        else:
            f.readline()
        symbols = []
        coords = []
        while True:
            l = f.readline().split()
            if 'end' in l:
                break
            symbols.append(l[0])
            coords.append([float(x) for x in l[1:4]])
        if pbc:
            image = Atoms(symbols, positions=coords, cell=cell, pbc=pbc)
        else:
            image = Atoms(symbols, positions=coords, pbc=pbc)
        images.append(image)
        l = f.readline()
        
                
    if not index:
        return images
    else:
        return images[index]

    




from ase.calculators.octopus import Octopus
from ase.build import bulk

system = bulk('Si', orthorhombic=True)

calc = Octopus(label='silicon',
               Spacing=0.25,
               KPointsGrid=[[4, 4, 4]],
               KPointsUseSymmetries=True,
               Output='dos + density + potential',
               OutputFormat='xcrysden',
               DosGamma=0.1)

system.set_calculator(calc)
system.get_potential_energy()



"""Check that QE calculation can run."""

from ase.build import bulk
from ase.calculators.espresso import Espresso

# Default pseudos can go in ~/espresso/pseudo
# Get these from SSSP http://materialscloud.org/sssp/
#PSEUDO = {'Si': 'Si.pbe-n-rrkjus_psl.1.0.0.UPF'}

# I am updating this so we use a file which is provided by the
# Debian/Ubuntu package for QE.  Then it works out of the
# box, which is good for testing.  --askhl
PSEUDO = {'Si': 'Si.rel-pbe-rrkj.UPF'}
# /usr/share/espresso/pseudo/Si.rel-pbe-rrkj.UPF

# Don't forget to
# export ASE_ESPRESSO_COMMAND="mpirun -n 4 $HOME/Compile/q-e/bin/pw.x -in PREFIX.pwi > PREFIX.pwo"
# export ESPRESSO_PSEUDO="/path/to/pseudos"

def main():
    silicon = bulk('Si')
    calc = Espresso(pseudopotentials=PSEUDO,
                    ecutwfc=50.0)
    silicon.set_calculator(calc)
    silicon.get_potential_energy()

    assert calc.get_fermi_level() is not None
    assert calc.get_ibz_k_points() is not None
    assert calc.get_eigenvalues(spin=0, kpt=0) is not None
    assert calc.get_number_of_spins() is not None
    assert calc.get_k_point_weights() is not None

main()




import numpy as np

from ase.calculators.eam import EAM

from ase.test.eam_pot import Pt_u3
from ase.build import fcc111
from io import StringIO

eam = EAM(potential=StringIO(Pt_u3), form='eam', elements=['Pt'])
slab = fcc111('Pt', size=(4, 4, 2), vacuum=10.0)
slab.set_calculator(eam)

assert( abs(-164.277599313 - slab.get_potential_energy()) < 1E-8 )
assert( abs(6.36379627645 - np.linalg.norm(slab.get_forces()))  < 1E-8 )



# -*- coding: utf-8 -*-
import pickle
import sys
import threading
from math import sqrt

import numpy as np

import ase.parallel as mpi
from ase.build import minimize_rotation_and_translation
from ase.calculators.calculator import Calculator
from ase.calculators.singlepoint import SinglePointCalculator
from ase.io import read
from ase.optimize import MDMin
from ase.geometry import find_mic
from ase.utils import basestring


class NEB:
    def __init__(self, images, k=0.1, fmax=0.05, climb=False, parallel=False,
                 remove_rotation_and_translation=False, world=None,
                 method='aseneb', dynamic_relaxation=False, scale_fmax=0.):
        """Nudged elastic band.

        Paper I:

            G. Henkelman and H. Jonsson, Chem. Phys, 113, 9978 (2000).
            https://doi.org/10.1063/1.1323224

        Paper II:

            G. Henkelman, B. P. Uberuaga, and H. Jonsson, Chem. Phys,
            113, 9901 (2000).
            https://doi.org/10.1063/1.1329672

        Paper III:

            E. L. Kolsbjerg, M. N. Groves, and B. Hammer, J. Chem. Phys,
            145, 094107 (2016)
            https://doi.org/10.1063/1.4961868

        images: list of Atoms objects
            Images defining path from initial to final state.
        k: float or list of floats
            Spring constant(s) in eV/Ang.  One number or one for each spring.
        climb: bool
            Use a climbing image (default is no climbing image).
        parallel: bool
            Distribute images over processors.
        remove_rotation_and_translation: bool
            TRUE actives NEB-TR for removing translation and
            rotation during NEB. By default applied non-periodic
            systems
        dynamic_relaxation: bool
            TRUE calculates the norm of the forces acting on each image
            in the band. An image is optimized only if its norm is above
            the convergence criterion. The list fmax_images is updated
            every force call; if a previously converged image goes out
            of tolerance (due to spring adjustments between the image
            and its neighbors), it will be optimized again. This routine
            can speed up calculations if convergence is non-uniform.
            Convergence criterion should be the same as that given to
            the optimizer. Not efficient when parallelizing over images.
        scale_fmax: float
            Scale convergence criteria along band based on the distance
            between a state and the state with the highest potential energy.
        method: string of method
            Choice betweeen three method:

            * aseneb: standard ase NEB implementation
            * improvedtangent: Paper I NEB implementation
            * eb: Paper III full spring force implementation
        """
        self.images = images
        self.climb = climb
        self.parallel = parallel
        self.natoms = len(images[0])
        pbc = images[0].pbc
        atomic_numbers = images[0].get_atomic_numbers()
        for img in images:
            if len(img) != self.natoms:
                raise ValueError('Images have different numbers of atoms')
            if (pbc != img.pbc).any():
                raise ValueError('Images have different boundary conditions')
            if (atomic_numbers != img.get_atomic_numbers()).any():
                raise ValueError('Images have atoms in different orders')
        self.nimages = len(images)
        self.emax = np.nan

        self.remove_rotation_and_translation = remove_rotation_and_translation
        self.dynamic_relaxation = dynamic_relaxation
        self.fmax = fmax
        self.scale_fmax = scale_fmax
        if not self.dynamic_relaxation and self.scale_fmax:
            msg = ('Scaled convergence criteria only implemented in series '
                   'with dynamic_relaxation.')
            raise ValueError(msg)

        if method in ['aseneb', 'eb', 'improvedtangent']:
            self.method = method
        else:
            raise NotImplementedError(method)

        if isinstance(k, (float, int)):
            k = [k] * (self.nimages - 1)
        self.k = list(k)

        if world is None:
            world = mpi.world
        self.world = world

        if parallel:
            assert world.size == 1 or world.size % (self.nimages - 2) == 0

        self.real_forces = None  # ndarray of shape (nimages, natom, 3)
        self.energies = None  # ndarray of shape (nimages,)

    def interpolate(self, method='linear', mic=False):
        if self.remove_rotation_and_translation:
            minimize_rotation_and_translation(self.images[0], self.images[-1])

        interpolate(self.images, mic)

        if method == 'idpp':
            self.idpp_interpolate(traj=None, log=None, mic=mic)

    def idpp_interpolate(self, traj='idpp.traj', log='idpp.log', fmax=0.1,
                         optimizer=MDMin, mic=False, steps=100):
        d1 = self.images[0].get_all_distances(mic=mic)
        d2 = self.images[-1].get_all_distances(mic=mic)
        d = (d2 - d1) / (self.nimages - 1)
        old = []
        for i, image in enumerate(self.images):
            old.append(image.calc)
            image.calc = IDPP(d1 + i * d, mic=mic)
        opt = optimizer(self, trajectory=traj, logfile=log)
        # BFGS was originally used by the paper, but testing shows that
        # MDMin results in nearly the same results in 3-4 orders of magnitude
        # less time. Known working optimizers = BFGS, MDMin, FIRE, HessLBFGS
        # Internal testing shows BFGS is only needed in situations where MDMIN
        # cannot converge easily and tends to be obvious on inspection.
        #
        # askhl: 3-4 orders of magnitude difference cannot possibly be
        # true unless something is actually broken.  Should it not be
        # "3-4 times"?
        opt.run(fmax=fmax, steps=steps)
        for image, calc in zip(self.images, old):
            image.calc = calc

    def get_positions(self):
        positions = np.empty(((self.nimages - 2) * self.natoms, 3))
        n1 = 0
        for image in self.images[1:-1]:
            n2 = n1 + self.natoms
            positions[n1:n2] = image.get_positions()
            n1 = n2
        return positions

    def set_positions(self, positions):
        n1 = 0
        for i, image in enumerate(self.images[1:-1]):
            if self.dynamic_relaxation:
                if self.parallel:
                    msg = ('Dynamic relaxation does not work efficiently '
                           'when parallelizing over images. Try AutoNEB '
                           'routine for freezing images in parallel.')
                    raise ValueError(msg)
                else:
                    forces_dyn = self.get_fmax_all(self.images)
                    if forces_dyn[i] < self.fmax:
                        n1 += self.natoms
                    else:
                        n2 = n1 + self.natoms
                        image.set_positions(positions[n1:n2])
                        n1 = n2
            else:
                n2 = n1 + self.natoms
                image.set_positions(positions[n1:n2])
                n1 = n2

    def get_fmax_all(self, images):
        n = self.natoms
        f_i = self.get_forces()
        fmax_images = []
        for i in range(self.nimages-2):
            n1 = n * i
            n2 = n + n * i
            fmax_images.append(np.sqrt((f_i[n1:n2]**2).sum(axis=1)).max())
        return fmax_images

    def get_forces(self):
        """Evaluate and return the forces."""
        images = self.images

        calculators = [image.calc for image in images
                       if image.calc is not None]
        if len(set(calculators)) != len(calculators):
            msg = ('One or more NEB images share the same calculator.  '
                   'Each image must have its own calculator.  '
                   'You may wish to use the ase.neb.SingleCalculatorNEB '
                   'class instead, although using separate calculators '
                   'is recommended.')
            raise ValueError(msg)

        forces = np.empty(((self.nimages - 2), self.natoms, 3))
        energies = np.empty(self.nimages)

        if self.remove_rotation_and_translation:
            # Remove translation and rotation between
            # images before computing forces:
            for i in range(1, self.nimages):
                minimize_rotation_and_translation(images[i - 1], images[i])

        if self.method != 'aseneb':
            energies[0] = images[0].get_potential_energy()
            energies[-1] = images[-1].get_potential_energy()

        if not self.parallel:
            # Do all images - one at a time:
            for i in range(1, self.nimages - 1):
                energies[i] = images[i].get_potential_energy()
                forces[i - 1] = images[i].get_forces()
        elif self.world.size == 1:
            def run(image, energies, forces):
                energies[:] = image.get_potential_energy()
                forces[:] = image.get_forces()
            threads = [threading.Thread(target=run,
                                        args=(images[i],
                                              energies[i:i + 1],
                                              forces[i - 1:i]))
                       for i in range(1, self.nimages - 1)]
            for thread in threads:
                thread.start()
            for thread in threads:
                thread.join()
        else:
            # Parallelize over images:
            i = self.world.rank * (self.nimages - 2) // self.world.size + 1
            try:
                energies[i] = images[i].get_potential_energy()
                forces[i - 1] = images[i].get_forces()
            except Exception:
                # Make sure other images also fail:
                error = self.world.sum(1.0)
                raise
            else:
                error = self.world.sum(0.0)
                if error:
                    raise RuntimeError('Parallel NEB failed!')

            for i in range(1, self.nimages - 1):
                root = (i - 1) * self.world.size // (self.nimages - 2)
                self.world.broadcast(energies[i:i + 1], root)
                self.world.broadcast(forces[i - 1], root)

        # Save for later use in iterimages:
        self.energies = energies
        self.real_forces = np.zeros((self.nimages, self.natoms, 3))
        self.real_forces[1:-1] = forces

        self.imax = 1 + np.argsort(energies[1:-1])[-1]
        self.emax = energies[self.imax]

        t1 = find_mic(images[1].get_positions() -
                      images[0].get_positions(),
                      images[0].get_cell(), images[0].pbc)[0]

        if self.method == 'eb':
            beeline = (images[self.nimages - 1].get_positions() -
                       images[0].get_positions())
            beelinelength = np.linalg.norm(beeline)
            eqlength = beelinelength / (self.nimages - 1)

        nt1 = np.linalg.norm(t1)

        for i in range(1, self.nimages - 1):
            t2 = find_mic(images[i + 1].get_positions() -
                          images[i].get_positions(),
                          images[i].get_cell(), images[i].pbc)[0]
            nt2 = np.linalg.norm(t2)

            if self.method == 'eb':
                # Tangents are bisections of spring-directions
                # (formula C8 of paper III)
                tangent = t1 / nt1 + t2 / nt2
                # Normalize the tangent vector
                tangent /= np.linalg.norm(tangent)
            elif self.method == 'improvedtangent':
                # Tangents are improved according to formulas 8, 9, 10,
                # and 11 of paper I.
                if energies[i + 1] > energies[i] > energies[i - 1]:
                    tangent = t2.copy()
                elif energies[i + 1] < energies[i] < energies[i - 1]:
                    tangent = t1.copy()
                else:
                    deltavmax = max(abs(energies[i + 1] - energies[i]),
                                    abs(energies[i - 1] - energies[i]))
                    deltavmin = min(abs(energies[i + 1] - energies[i]),
                                    abs(energies[i - 1] - energies[i]))
                    if energies[i + 1] > energies[i - 1]:
                        tangent = t2 * deltavmax + t1 * deltavmin
                    else:
                        tangent = t2 * deltavmin + t1 * deltavmax
                # Normalize the tangent vector
                tangent /= np.linalg.norm(tangent)
            else:
                if i < self.imax:
                    tangent = t2
                elif i > self.imax:
                    tangent = t1
                else:
                    tangent = t1 + t2
                tt = np.vdot(tangent, tangent)

            f = forces[i - 1]
            ft = np.vdot(f, tangent)

            if i == self.imax and self.climb:
                # imax not affected by the spring forces. The full force
                # with component along the elestic band converted
                # (formula 5 of Paper II)
                if self.method == 'aseneb':
                    f -= 2 * ft / tt * tangent
                else:
                    f -= 2 * ft * tangent
            elif self.method == 'eb':
                f -= ft * tangent
                # Spring forces
                # (formula C1, C5, C6 and C7 of Paper III)
                f1 = -(nt1 - eqlength) * t1 / nt1 * self.k[i - 1]
                f2 = (nt2 - eqlength) * t2 / nt2 * self.k[i]
                if self.climb and abs(i - self.imax) == 1:
                    deltavmax = max(abs(energies[i + 1] - energies[i]),
                                    abs(energies[i - 1] - energies[i]))
                    deltavmin = min(abs(energies[i + 1] - energies[i]),
                                    abs(energies[i - 1] - energies[i]))
                    f += (f1 + f2) * deltavmin / deltavmax
                else:
                    f += f1 + f2
            elif self.method == 'improvedtangent':
                f -= ft * tangent
                # Improved parallel spring force (formula 12 of paper I)
                f += (nt2 * self.k[i] - nt1 * self.k[i - 1]) * tangent
            else:
                f -= ft / tt * tangent
                f -= np.vdot(t1 * self.k[i - 1] -
                             t2 * self.k[i], tangent) / tt * tangent

            t1 = t2
            nt1 = nt2

            if self.dynamic_relaxation:
                n = self.natoms
                k = i - 1
                n1 = n * k
                n2 = n1 + n
                force_i = np.sqrt((forces.reshape((-1, 3))[n1:n2]**2.)
                                  .sum(axis=1)).max()

                n1_imax = (self.imax - 1) * n
                positions = self.get_positions()
                pos_imax = positions[n1_imax:n1_imax + n]
                rel_pos = np.sqrt(((positions[n1:n2] - pos_imax)**2).sum())

                if force_i < self.fmax * (1 + rel_pos * self.scale_fmax):
                    if k == self.imax - 1:
                        pass
                    else:
                        forces[k, :, :] = np.zeros((1, self.natoms, 3))
        return forces.reshape((-1, 3))

    def get_potential_energy(self, force_consistent=False):
        """Return the maximum potential energy along the band.
        Note that the force_consistent keyword is ignored and is only
        present for compatibility with ase.Atoms.get_potential_energy."""
        return self.emax

    def __len__(self):
        # Corresponds to number of optimizable degrees of freedom, i.e.
        # virtual atom count for the optimization algorithm.
        return (self.nimages - 2) * self.natoms

    def iterimages(self):
        # Allows trajectory to convert NEB into several images
        if not self.parallel or self.world.size == 1:
            for atoms in self.images:
                yield atoms
            return

        for i, atoms in enumerate(self.images):
            if i == 0 or i == self.nimages - 1:
                yield atoms
            else:
                atoms = atoms.copy()
                atoms.calc = SinglePointCalculator(energy=self.energies[i],
                                                   forces=self.real_forces[i],
                                                   atoms=atoms)
                yield atoms


class IDPP(Calculator):
    """Image dependent pair potential.

    See:
        Improved initial guess for minimum energy path calculations.
        Søren Smidstrup, Andreas Pedersen, Kurt Stokbro and Hannes Jónsson
        Chem. Phys. 140, 214106 (2014)
    """

    implemented_properties = ['energy', 'forces']

    def __init__(self, target, mic):
        Calculator.__init__(self)
        self.target = target
        self.mic = mic

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)

        P = atoms.get_positions()
        d = []
        D = []
        for p in P:
            Di = P - p
            if self.mic:
                Di, di = find_mic(Di, atoms.get_cell(), atoms.get_pbc())
            else:
                di = np.sqrt((Di**2).sum(1))
            d.append(di)
            D.append(Di)
        d = np.array(d)
        D = np.array(D)

        dd = d - self.target
        d.ravel()[::len(d) + 1] = 1  # avoid dividing by zero
        d4 = d**4
        e = 0.5 * (dd**2 / d4).sum()
        f = -2 * ((dd * (1 - 2 * dd / d) / d**5)[..., np.newaxis] * D).sum(0)
        self.results = {'energy': e, 'forces': f}


class SingleCalculatorNEB(NEB):
    def __init__(self, images, k=0.1, climb=False):
        if isinstance(images, basestring):
            # this is a filename
            images = read(images)

        NEB.__init__(self, images, k, climb, False)
        self.calculators = [None] * self.nimages
        self.energies_ok = False
        self.first = True

    def interpolate(self, initial=0, final=-1, mic=False):
        """Interpolate linearly between initial and final images."""
        if final < 0:
            final = self.nimages + final
        n = final - initial
        pos1 = self.images[initial].get_positions()
        pos2 = self.images[final].get_positions()
        dist = (pos2 - pos1)
        if mic:
            cell = self.images[initial].get_cell()
            assert((cell == self.images[final].get_cell()).all())
            pbc = self.images[initial].get_pbc()
            assert((pbc == self.images[final].get_pbc()).all())
            dist, D_len = find_mic(dist, cell, pbc)
        dist /= n
        for i in range(1, n):
            self.images[initial + i].set_positions(pos1 + i * dist)

    def refine(self, steps=1, begin=0, end=-1, mic=False):
        """Refine the NEB trajectory."""
        if end < 0:
            end = self.nimages + end
        j = begin
        n = end - begin
        for i in range(n):
            for k in range(steps):
                self.images.insert(j + 1, self.images[j].copy())
                self.calculators.insert(j + 1, None)
            self.k[j:j + 1] = [self.k[j] * (steps + 1)] * (steps + 1)
            self.nimages = len(self.images)
            self.interpolate(j, j + steps + 1, mic=mic)
            j += steps + 1

    def set_positions(self, positions):
        # new positions -> new forces
        if self.energies_ok:
            # restore calculators
            self.set_calculators(self.calculators[1:-1])
        NEB.set_positions(self, positions)

    def get_calculators(self):
        """Return the original calculators."""
        calculators = []
        for i, image in enumerate(self.images):
            if self.calculators[i] is None:
                calculators.append(image.get_calculator())
            else:
                calculators.append(self.calculators[i])
        return calculators

    def set_calculators(self, calculators):
        """Set new calculators to the images."""
        self.energies_ok = False
        self.first = True

        if not isinstance(calculators, list):
            calculators = [calculators] * self.nimages

        n = len(calculators)
        if n == self.nimages:
            for i in range(self.nimages):
                self.images[i].set_calculator(calculators[i])
        elif n == self.nimages - 2:
            for i in range(1, self.nimages - 1):
                self.images[i].set_calculator(calculators[i - 1])
        else:
            raise RuntimeError(
                'len(calculators)=%d does not fit to len(images)=%d'
                % (n, self.nimages))

    def get_energies_and_forces(self):
        """Evaluate energies and forces and hide the calculators"""
        if self.energies_ok:
            return

        self.emax = -1.e32

        def calculate_and_hide(i):
            image = self.images[i]
            calc = image.get_calculator()
            if self.calculators[i] is None:
                self.calculators[i] = calc
            if calc is not None:
                if not isinstance(calc, SinglePointCalculator):
                    self.images[i].set_calculator(
                        SinglePointCalculator(
                            image,
                            energy=image.get_potential_energy(
                                apply_constraint=False),
                            forces=image.get_forces(apply_constraint=False)))
                self.emax = min(self.emax, image.get_potential_energy())

        if self.first:
            calculate_and_hide(0)

        # Do all images - one at a time:
        for i in range(1, self.nimages - 1):
            calculate_and_hide(i)

        if self.first:
            calculate_and_hide(-1)
            self.first = False

        self.energies_ok = True

    def get_forces(self):
        self.get_energies_and_forces()
        return NEB.get_forces(self)

    def n(self):
        return self.nimages

    def write(self, filename):
        from ase.io.trajectory import Trajectory
        traj = Trajectory(filename, 'w', self)
        traj.write()
        traj.close()

    def __add__(self, other):
        for image in other:
            self.images.append(image)
        return self


def fit0(E, F, R, cell=None, pbc=None):
    """Constructs curve parameters from the NEB images."""
    E = np.array(E) - E[0]
    n = len(E)
    Efit = np.empty((n - 1) * 20 + 1)
    Sfit = np.empty((n - 1) * 20 + 1)

    s = [0]
    dR = np.zeros_like(R)
    for i in range(n):
        if i < n - 1:
            dR[i] = R[i + 1] - R[i]
            if cell is not None and pbc is not None:
                dR[i], _ = find_mic(dR[i], cell, pbc)
            s.append(s[i] + sqrt((dR[i]**2).sum()))
        else:
            dR[i] = R[i] - R[i - 1]
            if cell is not None and pbc is not None:
                dR[i], _ = find_mic(dR[i], cell, pbc)

    lines = []
    dEds0 = None
    for i in range(n):
        d = dR[i]
        if i == 0:
            ds = 0.5 * s[1]
        elif i == n - 1:
            ds = 0.5 * (s[-1] - s[-2])
        else:
            ds = 0.25 * (s[i + 1] - s[i - 1])

        d = d / sqrt((d**2).sum())
        dEds = -(F[i] * d).sum()
        x = np.linspace(s[i] - ds, s[i] + ds, 3)
        y = E[i] + dEds * (x - s[i])
        lines.append((x, y))

        if i > 0:
            s0 = s[i - 1]
            s1 = s[i]
            x = np.linspace(s0, s1, 20, endpoint=False)
            c = np.linalg.solve(np.array([(1, s0, s0**2, s0**3),
                                          (1, s1, s1**2, s1**3),
                                          (0, 1, 2 * s0, 3 * s0**2),
                                          (0, 1, 2 * s1, 3 * s1**2)]),
                                np.array([E[i - 1], E[i], dEds0, dEds]))
            y = c[0] + x * (c[1] + x * (c[2] + x * c[3]))
            Sfit[(i - 1) * 20:i * 20] = x
            Efit[(i - 1) * 20:i * 20] = y

        dEds0 = dEds

    Sfit[-1] = s[-1]
    Efit[-1] = E[-1]
    return s, E, Sfit, Efit, lines


class NEBTools:
    """Class to make many of the common tools for NEB analysis available to
    the user. Useful for scripting the output of many jobs. Initialize with
    list of images which make up a single band."""

    def __init__(self, images):
        self._images = images

    def get_barrier(self, fit=True, raw=False):
        """Returns the barrier estimate from the NEB, along with the
        Delta E of the elementary reaction. If fit=True, the barrier is
        estimated based on the interpolated fit to the images; if
        fit=False, the barrier is taken as the maximum-energy image
        without interpolation. Set raw=True to get the raw energy of the
        transition state instead of the forward barrier."""
        s, E, Sfit, Efit, lines = self.get_fit()
        dE = E[-1] - E[0]
        if fit:
            barrier = max(Efit)
        else:
            barrier = max(E)
        if raw:
            barrier += self._images[0].get_potential_energy()
        return barrier, dE

    def plot_band(self, ax=None):
        """Plots the NEB band on matplotlib axes object 'ax'. If ax=None
        returns a new figure object."""
        ax = plot_band_from_fit(*self.get_fit(), ax=ax)
        return ax.figure

    def get_fmax(self, **kwargs):
        """Returns fmax, as used by optimizers with NEB."""
        neb = NEB(self._images, **kwargs)
        forces = neb.get_forces()
        return np.sqrt((forces**2).sum(axis=1).max())

    def get_fit(self):
        """Returns the parameters for fitting images to band."""
        images = self._images
        R = [atoms.positions for atoms in images]
        E = [atoms.get_potential_energy() for atoms in images]
        F = [atoms.get_forces() for atoms in images]
        A = images[0].cell
        pbc = images[0].pbc
        s, E, Sfit, Efit, lines = fit0(E, F, R, A, pbc)
        return s, E, Sfit, Efit, lines


def plot_band_from_fit(s, E, Sfit, Efit, lines, ax=None):
    if ax is None:
        import matplotlib.pyplot as plt
        ax = plt.gca()

    ax.plot(s, E, 'o')
    for x, y in lines:
        ax.plot(x, y, '-g')
    ax.plot(Sfit, Efit, 'k-')
    ax.set_xlabel(r'path [$\AA$]')
    ax.set_ylabel('energy [eV]')
    Ef = max(Efit) - E[0]
    Er = max(Efit) - E[-1]
    dE = E[-1] - E[0]
    ax.set_title('$E_\\mathrm{f} \\approx$ %.3f eV; '
                 '$E_\\mathrm{r} \\approx$ %.3f eV; '
                 '$\\Delta E$ = %.3f eV'
                 % (Ef, Er, dE))
    return ax


NEBtools = NEBTools  # backwards compatibility


def interpolate(images, mic=False):
    """Given a list of images, linearly interpolate the positions of the
    interior images."""
    pos1 = images[0].get_positions()
    pos2 = images[-1].get_positions()
    d = pos2 - pos1
    if mic:
        d = find_mic(d, images[0].get_cell(), images[0].pbc)[0]
    d /= (len(images) - 1.0)
    for i in range(1, len(images) - 1):
        images[i].set_positions(pos1 + i * d)


if __name__ == '__main__':
    # This stuff is used by ASE's GUI
    import matplotlib.pyplot as plt
    fit = pickle.load(sys.stdin)
    plot_band_from_fit(*fit)
    plt.show()



"""Tests for class ExternalForce in ase/constraints.py"""
from ase import Atoms
from ase.constraints import ExternalForce, FixBondLength
from ase.optimize import FIRE
from ase.calculators.emt import EMT
from numpy.linalg import norm


f_ext = 0.2

atom1 = 0
atom2 = 1

atom3 = 2

fmax = 0.001

atoms = Atoms('H3', positions=[(0, 0, 0), (0.751, 0, 0), (0, 1., 0)])
atoms.set_calculator(EMT())

# Without external force
opt = FIRE(atoms)
opt.run(fmax=fmax)
dist1 = atoms.get_distance(atom1, atom2)

# With external force
con1 = ExternalForce(atom1, atom2, f_ext)
atoms.set_constraint(con1)
opt = FIRE(atoms)
opt.run(fmax=fmax)
dist2 = atoms.get_distance(atom1, atom2)
# Distance should increase due to the external force
assert dist2 > dist1

# Combine ExternalForce with FixBondLength

# Fix the bond on which the force acts
con2 = FixBondLength(atom1, atom2)
# ExternalForce constraint at the beginning of the list!!!
atoms.set_constraint([con1, con2])
opt = FIRE(atoms)
opt.run(fmax=fmax)
f_con = con2.constraint_forces

# It was already optimized with this external force, therefore
# the constraint force should be almost zero
assert norm(f_con[0]) <= fmax

# To get the complete constraint force (with external force),
# use only the FixBondLength constraint, after the optimization with
# ExternalForce
atoms.set_constraint(con2)
opt = FIRE(atoms)
opt.run(fmax=fmax)
f_con = con2.constraint_forces[0]
assert round(norm(f_con), 2) == round(abs(f_ext), 2)

# Fix another bond and incrase the external force
f_ext *= 2
con1 = ExternalForce(atom1, atom2, f_ext)
d1 = atoms.get_distance(atom1, atom3)
con2 = FixBondLength(atom1, atom3)
# ExternalForce constraint at the beginning of the list!!!
atoms.set_constraint([con1, con2])
opt = FIRE(atoms)
opt.run(fmax=fmax)
d2 = atoms.get_distance(atom1, atom3)
# Fixed distance should not change
assert round(d1, 5) == round(d2, 5)



from gpaw import restart
from ase.dft import Wannier

atoms, calc = restart('benzene.gpw', txt=None)

# Make wannier functions of occupied space only
wan = Wannier(nwannier=15, calc=calc)
wan.localize()
for i in range(wan.nwannier):
    wan.write_cube(i, 'benzene15_%i.cube' % i)

# Make wannier functions using (three) extra degrees of freedom.
wan = Wannier(nwannier=18, calc=calc, fixedstates=15)
wan.localize()
wan.save('wan18.pickle')
for i in range(wan.nwannier):
    wan.write_cube(i, 'benzene18_%i.cube' % i)



from ase.build import niggli_reduce
from ase.calculators.singlepoint import SinglePointCalculator
from ase.optimize import FIRE
from ase.constraints import ExpCellFilter
from ase.ga import set_raw_score
from asap3 import EMT


def finalize(atoms, energy=None, forces=None, stress=None):
    # Finalizes the atoms by attaching a SinglePointCalculator
    # and setting the raw score as the negative of the total energy
    atoms.wrap()
    calc = SinglePointCalculator(atoms, energy=energy, forces=forces,
                                 stress=stress)
    atoms.set_calculator(calc)
    raw_score = -atoms.get_potential_energy()
    set_raw_score(atoms, raw_score)


def relax(atoms, cellbounds=None):
    # Performs a variable-cell relaxation of the structure
    calc = EMT()
    atoms.set_calculator(calc)

    converged = False
    niter = 0
    while not converged and niter < 10:
        if cellbounds is not None:
            cell = atoms.get_cell()
            if not cellbounds.is_within_bounds(cell):
                niggli_reduce(atoms)
            cell = atoms.get_cell()
            if not cellbounds.is_within_bounds(cell):
                # Niggli reduction did not bring the unit cell
                # within the specified bounds; this candidate should
                # be discarded so we set an absurdly high energy
                finalize(atoms, 1e9)
                return

        ecf = ExpCellFilter(atoms)
        dyn = FIRE(ecf, maxmove=0.2, logfile=None, trajectory=None)
        dyn.run(fmax=1e-3, steps=100)

        converged = dyn.converged()
        niter += 1

    dyn = FIRE(atoms, maxmove=0.2, logfile=None, trajectory=None)
    dyn.run(fmax=1e-2, steps=100)

    e = atoms.get_potential_energy()
    f = atoms.get_forces()
    s = atoms.get_stress()
    finalize(atoms, energy=e, forces=f, stress=s)



# creates: bztable.rst
# creates: 00.CUB.svg 01.FCC.svg 02.BCC.svg 03.TET.svg 04.BCT1.svg
# creates: 05.BCT2.svg 06.ORC.svg 07.ORCF1.svg 08.ORCF2.svg 09.ORCF3.svg
# creates: 10.ORCI.svg 11.ORCC.svg 12.HEX.svg 13.RHL1.svg 14.RHL2.svg
# creates: 15.MCL.svg 16.MCLC1.svg 17.MCLC3.svg 18.MCLC5.svg 19.TRI1a.svg
# creates: 20.TRI1b.svg 21.TRI2a.svg 22.TRI2b.svg
# creates: 23.OBL.svg 24.RECT.svg 25.CRECT.svg 26.HEX2D.svg 27.SQR.svg

from ase.lattice import all_variants


header = """\

Brillouin zone data
-------------------

.. list-table::
    :widths: 10 15 45
"""


entry = """\
    * - {name} ({longname})
      - {bandpath}
      - .. image:: {fname}
            :width: 40 %
"""

with open('bztable.rst', 'w') as fd:
    print(header, file=fd)

    for i, lat in enumerate(all_variants(include_blunt_angles=False)):
        id = '{:02d}.{}'.format(i, lat.variant)
        imagefname = '{}.svg'.format(id)
        txt = entry.format(name=lat.variant,
                           longname=lat.longname,
                           bandpath=lat.bandpath().path,
                           fname=imagefname)
        print(txt, file=fd)
        ax = lat.plot_bz()
        fig = ax.get_figure()
        fig.savefig(imagefname, bbox_inches='tight')
        fig.clear()



# creates: atomic_radii.png
# encoding: utf-8

import numpy as np
import matplotlib.pyplot as plt
from ase.data.vdw import vdw_radii as vdw1
from ase.data.vdw_alvarez import vdw_radii as vdw2
from ase.data import covalent_radii, chemical_symbols

plt.grid(ls=':')
c1 = covalent_radii.copy()
c1[c1 < 0.2001] = np.nan  # Remove 'false' values which are all 0.2
plt.plot(vdw2, marker='.', label='vdw_radii [ase.data.vdw_alvarez]')
plt.plot(vdw1, marker='.', label='vdw_radii [ase.data.vdw]')
plt.plot(c1, marker='.', label='covalent_radii [ase.data]')
nobles = [2, 10, 18, 36, 54, 86]
plt.xticks(nobles, [chemical_symbols[Z] for Z in nobles])
plt.xlabel('Z')
plt.ylabel(u'radius [Å]')
plt.legend(loc='best')
plt.savefig('atomic_radii.png')



from ase.calculators.dftb import Dftb
from ase.optimize import QuasiNewton
from ase.io import write

from ase.build import molecule
test = molecule('H2O')
test.set_calculator(Dftb(label='h2o',
                         atoms=test,
                         Hamiltonian_MaxAngularMomentum_='',
                         Hamiltonian_MaxAngularMomentum_O='"p"',
                         Hamiltonian_MaxAngularMomentum_H='"s"',
                         ))

dyn = QuasiNewton(test, trajectory='test.traj')
dyn.run(fmax=0.01)
write('test.final.xyz', test)



from ase.data import atomic_numbers
from ase.ga.utilities import closest_distances_generator
from ase.ga.bulk_utilities import CellBounds
from ase.ga.bulk_startgenerator import StartGenerator
from ase.ga.data import PrepareDB

# Number of random initial structures to generate
N = 20

# Target cell volume for the initial structures, in angstrom^3
volume = 240.

# Specify the 'building blocks' from which the initial structures
# will be constructed. Here we take single Ag atoms as building
# blocks, 24 in total.
blocks = [('Ag', 24)]

# From these blocks we can determine the stoichiometry
# (as a list of atomic numbers)
stoichiometry = [atomic_numbers[atom] for atom, count in blocks
                 for _ in range(count)]

# Generate a dictionary with the closest allowed interatomic distances
atom_numbers = list(set(stoichiometry))
blmin = closest_distances_generator(atom_numbers=atom_numbers,
                                    ratio_of_covalent_radii=0.5)

# Specify reasonable bounds on the minimal and maximal
# cell vector lengths (in angstrom) and angles (in degrees)
cellbounds = CellBounds(bounds={'phi': [35, 145],
                                'chi': [35, 145],
                                'psi': [35, 145],
                                'a': [3, 50], 'b': [3, 50], 'c': [3, 50]})

# Choose an (optional) 'cell splitting' scheme which basically
# controls the level of translational symmetry (within the unit
# cell) of the randomly generated structures. Here a 1:1 ratio
# of splitting factors 2 and 1 is used:
splits = {(2,): 1, (1,): 1}
# There will hence be a 50% probability that a candidate
# is constructed by repeating an randomly generated Ag12
# structure along a randomly chosen axis. In the other 50%
# of cases, no cell cell splitting will be applied.

# Initialize the random structure generator
sg = StartGenerator(blocks, blmin, volume, cellbounds=cellbounds,
                    splits=splits)

# Create the database
da = PrepareDB(db_file_name='gadb.db',
               stoichiometry=stoichiometry)

# Generate N random structures
# and add them to the database
for i in range(N):
    a = sg.get_new_candidate()
    da.add_unrelaxed_candidate(a)
